"""
SOQL Builder API - Constrained and safe for free tier
"""

from fastapi import APIRouter, Depends, HTTPException, Request, Query
from typing import Dict, Any, List, Optional
from pydantic import BaseModel
import json
import re

from .deps import require_api_key
from ...services.soql_query_policy import (
    canonical_safe_object_name,
    safe_field_names_for_object,
)

router = APIRouter(prefix="/salesforce/query-builder", tags=["SOQL Builder"])


class SOQLValidationRequest(BaseModel):
    soql: str
    validate_only: bool = True


class SOQLValidationResponse(BaseModel):
    valid: bool
    warnings: List[str] = []
    estimated_rows: Optional[int] = None
    api_calls_remaining: int
    projected_api_calls: int
    will_hit_limit: bool = False


@router.get("/objects")
async def list_queryable_objects(api_ctx=Depends(require_api_key)):
    """
    List queryable Salesforce objects (filtered for safety).
    Only returns: queryable=true AND supportsFeed=false
    """
    from ...services.salesforce_gateway import SalesforceGateway
    from ...db import get_session

    account_id = api_ctx.account_id
    if not account_id:
        raise HTTPException(401, "Missing account_id")

    async for db in get_session():
        try:
            gateway = await SalesforceGateway.for_tenant(str(account_id), db)

            # Get all objects
            all_objects = await gateway.describe_global()

            # Filter to safe, queryable objects
            safe_objects = [
                {
                    "name": obj["name"],
                    "label": obj["label"],
                    "labelPlural": obj.get("labelPlural", obj["label"]),
                    "keyPrefix": obj.get("keyPrefix"),
                }
                for obj in all_objects.get("sobjects", [])
                if obj.get("queryable")
                and not obj.get("supportsFeed", False)  # Exclude feed objects
                and not obj["name"].endswith("__Feed")
                and not obj["name"].endswith("__History")
                and not obj["name"].endswith("__Share")
            ]

            # Sort by label
            safe_objects.sort(key=lambda x: x["label"])

            return {"ok": True, "data": {"objects": safe_objects}}
        except Exception as e:
            raise HTTPException(500, f"Failed to fetch objects: {e}")


@router.api_route("/fields", methods=["GET", "POST"])
async def get_object_fields(
    request: Request,
    object_name: Optional[str] = Query(None, alias="object"),
    api_ctx=Depends(require_api_key),
):
    """
    Get fields for a specific object with metadata for safe queries.
    Returns: sortable, filterable, aggregatable flags
    """
    from ...services.salesforce_gateway import SalesforceGateway
    from ...db import get_session

    account_id = api_ctx.account_id
    if not account_id:
        raise HTTPException(401, "Missing account_id")

    if not object_name and request.method == "POST":
        if request.method == "POST":
            try:
                body = await request.json()
            except Exception:
                body = {}
            object_name = body.get("object") or body.get("object_name")
    if not object_name:
        raise HTTPException(400, "Missing object parameter")

    async for db in get_session():
        try:
            gateway = await SalesforceGateway.for_tenant(str(account_id), db)

            # Get object describe
            describe = await gateway.describe_object(object_name)
            resolved_object_name = describe.get("name") or object_name
            allowlist_object = canonical_safe_object_name(
                resolved_object_name
            ) or canonical_safe_object_name(object_name)
            allowlisted_fields = (
                safe_field_names_for_object(allowlist_object)
                if allowlist_object
                else None
            )

            # Process fields with safety metadata and query-execution policy.
            fields = []
            hidden_by_policy = 0
            hidden_by_schema = 0
            for field in describe.get("fields", []):
                field_name = str(field.get("name") or "").strip()
                if not field_name:
                    continue
                field_type = str(field.get("type") or "").strip()
                field_type_lower = field_type.lower()
                if field_type_lower in {"address", "location"} or bool(
                    field.get("deprecatedAndHidden")
                ):
                    hidden_by_schema += 1
                    continue
                if not field.get("queryable", True) or not field.get(
                    "accessible", True
                ):
                    hidden_by_schema += 1
                    continue
                if (
                    allowlisted_fields is not None
                    and field_name.lower() not in allowlisted_fields
                ):
                    hidden_by_policy += 1
                    continue

                field_info = {
                    "name": field_name,
                    "label": field.get("label") or field_name,
                    "type": field_type,
                    "sortable": field.get("sortable", False),
                    "filterable": field.get("filterable", False),
                    "groupable": field.get("groupable", False),
                    "aggregatable": field.get("aggregatable", False),
                    "nillable": field.get("nillable", True),
                    "relationshipName": field.get("relationshipName"),
                    "referenceTo": field.get("referenceTo", []),
                    "length": field.get("length"),
                    "precision": field.get("precision"),
                    "scale": field.get("scale"),
                    "calculated": field.get("calculated", False),
                    "restrictedPicklist": field.get("restrictedPicklist", False),
                    "dependentPicklist": field.get("dependentPicklist", False),
                    "controllerName": field.get("controllerName"),
                    "helpText": field.get("inlineHelpText") or field.get("description"),
                }

                picklist_values = field.get("picklistValues") or []
                if picklist_values:
                    trimmed = []
                    for entry in picklist_values[:200]:
                        value = entry.get("value")
                        if value in (None, ""):
                            continue
                        trimmed.append(
                            {
                                "value": str(value),
                                "label": str(entry.get("label") or value),
                                "default": bool(entry.get("defaultValue")),
                            }
                        )
                    field_info["picklistValues"] = trimmed

                # Mark relationship fields
                if field.get("relationshipName"):
                    field_info["isRelationship"] = True
                    field_info["warning"] = (
                        "Relationship field - counts as parent lookup (use Join-Lite for complex joins)"
                    )

                fields.append(field_info)

            # Sort: standard fields first, then custom
            fields.sort(key=lambda x: (x["name"].endswith("__c"), x["label"]))

            return {
                "ok": True,
                "data": {
                    "object": resolved_object_name,
                    "fields": fields,
                    "totalFields": len(fields),
                    "hiddenByPolicy": hidden_by_policy,
                    "hiddenBySchema": hidden_by_schema,
                    "allowlistObject": allowlist_object,
                },
            }
        except Exception as e:
            import logging

            logger = logging.getLogger(__name__)
            logger.error(
                f"SOQL Builder fields error - account_id={account_id}, "
                f"object={object_name}, error={str(e)}",
                exc_info=True,
            )
            raise HTTPException(500, f"Failed to fetch fields: {e}")


@router.api_route("/field-values", methods=["GET", "POST"])
async def get_field_values(
    request: Request,
    object_name: Optional[str] = Query(None, alias="object"),
    field: Optional[str] = Query(None),
    limit: int = Query(10),
    api_ctx=Depends(require_api_key),
):
    """
    Return a de-duplicated sample of field values to help users build filters.
    """
    from ...services.salesforce_gateway import SalesforceGateway
    from ...db import get_session

    account_id = api_ctx.account_id
    if not account_id:
        raise HTTPException(401, "Missing account_id")

    if request.method == "POST":
        try:
            body = await request.json()
        except Exception:
            body = {}
        object_name = object_name or body.get("object") or body.get("object_name")
        field = field or body.get("field")
        if "limit" in body and body["limit"] is not None:
            try:
                limit = int(body["limit"])
            except (ValueError, TypeError):
                raise HTTPException(400, "Invalid limit parameter")

    if not object_name or not re.fullmatch(r"[A-Za-z0-9_]+", str(object_name)):
        raise HTTPException(400, "Invalid object name")
    if not field or not re.fullmatch(r"[A-Za-z0-9_.]+", str(field)):
        raise HTTPException(400, "Invalid field name")

    try:
        limit = int(limit or 10)
    except (ValueError, TypeError):
        raise HTTPException(400, "Invalid limit parameter")

    limit = max(1, min(limit, 25))

    async for db in get_session():
        try:
            gateway = await SalesforceGateway.for_tenant(str(account_id), db)
            describe = await gateway.describe_object(object_name)
            field_map = {f["name"]: f for f in describe.get("fields", [])}

            path_parts = field.split(".")
            is_relationship = len(path_parts) > 1
            leaf_field = path_parts[-1]

            related_object = None
            base_field_meta = None
            related_describe = None

            if is_relationship:
                relationship_name = path_parts[0]
                rel_field_meta = next(
                    (
                        f
                        for f in describe.get("fields", [])
                        if f.get("relationshipName") == relationship_name
                    ),
                    None,
                )
                if not rel_field_meta:
                    raise HTTPException(
                        400,
                        f"Relationship {relationship_name} not found on {object_name}",
                    )
                reference_to = rel_field_meta.get("referenceTo") or []
                if not reference_to:
                    raise HTTPException(
                        400, f"No reference object for relationship {relationship_name}"
                    )
                related_object = reference_to[0]
                related_describe = await gateway.describe_object(related_object)
                related_map = {f["name"]: f for f in related_describe.get("fields", [])}
                base_field_meta = related_map.get(leaf_field)
                if not base_field_meta:
                    raise HTTPException(
                        400, f"Field {leaf_field} not found on {related_object}"
                    )
            else:
                base_field_meta = field_map.get(field)
                if not base_field_meta:
                    raise HTTPException(
                        400, f"Field {field} not found on {object_name}"
                    )

            display_field_path = field
            reference_relationship_name: Optional[str] = None
            reference_targets = (
                base_field_meta.get("referenceTo", []) if base_field_meta else []
            )
            reference_multi_target = len(reference_targets) > 1

            if (
                not is_relationship
                and base_field_meta
                and base_field_meta.get("type") == "reference"
                and not reference_multi_target
            ):
                reference_relationship_name = base_field_meta.get("relationshipName")
                if reference_relationship_name:
                    display_field_path = f"{reference_relationship_name}.Name"
                    if not related_object and reference_targets:
                        related_object = reference_targets[0]
                else:
                    display_field_path = field
                if related_object and not related_describe:
                    related_describe = await gateway.describe_object(related_object)
            if is_relationship:
                filter_field = rel_field_meta.get("name") if rel_field_meta else None
            else:
                filter_field = (
                    field if base_field_meta.get("filterable", False) else None
                )

            def _build_allowlist() -> Dict[str, Any]:
                objects = {object_name}
                fields_map: Dict[str, set[str]] = {}
                base_key = path_parts[0] if is_relationship else leaf_field
                fields_map.setdefault(object_name.lower(), set()).add(base_key)
                if filter_field:
                    fields_map.setdefault(object_name.lower(), set()).add(filter_field)
                if is_relationship and related_object:
                    objects.add(related_object)
                    fields_map.setdefault(related_object.lower(), set()).add(leaf_field)
                if (
                    not is_relationship
                    and reference_relationship_name
                    and not reference_multi_target
                ):
                    fields_map.setdefault(object_name.lower(), set()).add(
                        reference_relationship_name
                    )
                if (
                    related_object
                    and display_field_path != field
                    and not reference_multi_target
                ):
                    objects.add(related_object)
                    fields_map.setdefault(related_object.lower(), set()).add("Name")
                return {
                    "objects": list(objects),
                    "fields": {
                        obj: sorted(list(vals)) for obj, vals in fields_map.items()
                    },
                }

            plan_allowlist = _build_allowlist()

            sample_limit = limit * 5

            async def _run_sample(where_clause: str) -> List[Dict[str, Any]]:
                clause = f" WHERE {where_clause}" if where_clause else ""
                select_fields: List[str] = []
                for candidate in (field, display_field_path):
                    if candidate and candidate not in select_fields:
                        select_fields.append(candidate)
                soql = f"SELECT {', '.join(select_fields)} FROM {object_name}{clause} LIMIT {sample_limit}"
                result = await gateway.soql(
                    soql,
                    plan_allowlist=plan_allowlist,
                    fls_guard=False,
                )
                return result.get("records", []) or []

            records = await _run_sample(
                f"{filter_field} != NULL" if filter_field else ""
            )
            if not records and filter_field:
                records = await _run_sample("")

            async def _run_grouped(where_clause: str) -> List[Dict[str, Any]]:
                clause = f" WHERE {where_clause}" if where_clause else ""
                soql = f"SELECT {field} FROM {object_name}{clause} GROUP BY {field} LIMIT {sample_limit}"
                result = await gateway.soql(
                    soql,
                    plan_allowlist=plan_allowlist,
                    fls_guard=False,
                )
                return result.get("records", []) or []

            can_group = bool(base_field_meta.get("groupable", False))
            # Grouping on relationship display fields is unsupported, so disable when lookups
            if display_field_path != field:
                can_group = False
            if is_relationship and base_field_meta is not None:
                can_group = can_group or bool(base_field_meta.get("groupable", False))

            if not records and can_group:
                try:
                    records = await _run_grouped(
                        f"{filter_field} != NULL" if filter_field else ""
                    )
                    if not records and filter_field:
                        records = await _run_grouped("")
                except Exception:
                    records = records or []

            def _extract_value(record: Dict[str, Any], path: str) -> Optional[Any]:
                current: Any = record
                for part in path.split("."):
                    if not isinstance(current, dict):
                        return None
                    current = current.get(part)
                    if current is None:
                        return None
                if isinstance(current, dict):
                    for key in ("Name", "name", "FullName", "Label", "value", "Id"):
                        if key in current and current[key] not in (None, ""):
                            return current[key]
                    return json.dumps(current, ensure_ascii=False)
                return current

            seen: set[str] = set()
            values: List[Dict[str, Any]] = []
            display_index: Dict[str, List[int]] = {}

            for record in records:
                extracted = _extract_value(record, field)
                if extracted in (None, ""):
                    continue
                raw_string = extracted if isinstance(extracted, str) else str(extracted)
                if not raw_string:
                    continue
                if raw_string in seen:
                    continue
                seen.add(raw_string)

                display_value = raw_string
                if display_field_path != field:
                    display_candidate = _extract_value(record, display_field_path)
                    if display_candidate not in (None, ""):
                        display_value = (
                            display_candidate
                            if isinstance(display_candidate, str)
                            else str(display_candidate)
                        )
                else:
                    if isinstance(extracted, str):
                        display_value = extracted
                    else:
                        display_value = str(extracted)

                raw_for_insert = (
                    extracted if display_field_path == field else raw_string
                )

                display_index.setdefault(display_value, []).append(len(values))

                entry: Dict[str, Any] = {"value": display_value, "raw": raw_for_insert}
                if display_field_path != field:
                    entry["id"] = raw_string

                values.append(entry)
                if len(values) >= limit:
                    break

            for label, idxs in display_index.items():
                if len(idxs) > 1:
                    for idx in idxs:
                        entry = values[idx]
                        identifier = entry.get("id") or entry.get("raw")
                        if not identifier:
                            continue
                        identifier_str = str(identifier)
                        short_id = (
                            f"{identifier_str[:6]}…{identifier_str[-4:]}"
                            if len(identifier_str) > 12
                            else identifier_str
                        )
                        entry["value"] = f"{label} ({short_id})"

            return {
                "ok": True,
                "data": {"values": values, "field": field, "object": object_name},
            }
        except HTTPException:
            raise
        except Exception as exc:  # noqa: BLE001
            raise HTTPException(500, f"Failed to fetch field values: {exc}") from exc


@router.post("/validate")
async def validate_soql(
    request: SOQLValidationRequest, api_ctx=Depends(require_api_key)
):
    """
    Validate SOQL and return warnings + estimates.
    Guards against expensive queries.
    """
    from ...services.salesforce_gateway import SalesforceGateway
    from ...db import get_session

    account_id = api_ctx.account_id
    if not account_id:
        raise HTTPException(401, "Missing account_id")

    soql = request.soql.strip()
    warnings = []
    estimated_rows = None

    # Parse SOQL for common issues
    soql_upper = soql.upper()

    # Check for WHERE clause on large objects
    large_objects = ["ACCOUNT", "CONTACT", "LEAD", "OPPORTUNITY", "CASE"]
    for obj in large_objects:
        if f"FROM {obj}" in soql_upper and "WHERE" not in soql_upper:
            warnings.append(
                f"No WHERE clause on {obj}. This may return many rows and use API calls."
            )

    # Check for LIMIT clause
    if "LIMIT" not in soql_upper:
        warnings.append(
            "No LIMIT clause. Consider adding LIMIT to control result size."
        )

    # Check for ORDER BY (recommended for pagination stability)
    if "LIMIT" in soql_upper and "ORDER BY" not in soql_upper:
        warnings.append(
            "LIMIT without ORDER BY may return inconsistent results across pages. Add ORDER BY Id."
        )

    # Check for child relationships (not supported in free tier)
    if "SELECT (" in soql:
        warnings.append(
            "Child relationship query detected. Use Join-Lite or upgrade to Pro for server-side joins."
        )

    async for db in get_session():
        try:
            gateway = await SalesforceGateway.for_tenant(str(account_id), db)

            # Resolve API limit headroom using live data when available,
            # otherwise fall back to the gateway's cached telemetry.
            def _coerce_int(value: Optional[Any]) -> Optional[int]:
                try:
                    return int(value)
                except (TypeError, ValueError):
                    return None

            limits_raw = None
            fetch_limits = getattr(gateway, "get_api_limits", None)
            if callable(fetch_limits):
                try:
                    limits_raw = await fetch_limits()
                except Exception:
                    # Silently fall back to cached values; live limits aren't always available.
                    limits_raw = None

            api_remaining = None
            api_max = None

            if isinstance(limits_raw, dict):
                daily_limits = limits_raw.get("DailyApiRequests") or limits_raw.get(
                    "daily_api_requests"
                )
                if isinstance(daily_limits, dict):
                    api_remaining = _coerce_int(
                        daily_limits.get("Remaining") or daily_limits.get("remaining")
                    )
                    api_max = _coerce_int(
                        daily_limits.get("Max") or daily_limits.get("max")
                    )
                    used = _coerce_int(
                        daily_limits.get("Used") or daily_limits.get("used")
                    )
                    if (
                        api_max is None
                        and api_remaining is not None
                        and used is not None
                    ):
                        api_max = api_remaining + used

            if api_max is None or api_remaining is None:
                parsed_usage = getattr(gateway, "_parsed_api_usage", None)
                if isinstance(parsed_usage, tuple) and len(parsed_usage) == 2:
                    used, total = parsed_usage
                    used_int = _coerce_int(used)
                    total_int = _coerce_int(total)
                    if total_int:
                        if api_max is None:
                            api_max = total_int
                        if api_remaining is None and used_int is not None:
                            api_remaining = max(total_int - used_int, 0)

            if api_remaining is None:
                cached_remaining = _coerce_int(
                    getattr(gateway, "_remaining_api_calls", None)
                )
                if cached_remaining is not None:
                    api_remaining = max(cached_remaining, 0)

            if api_max is None:
                api_max = 5000
            if api_remaining is None:
                api_remaining = max(int(api_max * 0.8), 1)

            # Estimate API calls needed (rough heuristic)
            projected_calls = 1  # Initial query
            if "LIMIT" in soql_upper:
                # Extract LIMIT value
                import re

                limit_match = re.search(r"LIMIT\s+(\d+)", soql_upper)
                if limit_match:
                    limit_val = int(limit_match.group(1))
                    # 2000 rows per API call
                    projected_calls = max(1, (limit_val // 2000) + 1)
            else:
                # Without LIMIT, assume potentially many calls
                projected_calls = 10
                warnings.append("Without LIMIT, query may use many API calls.")

            # Check if we'll hit 80% threshold
            will_hit_limit = (api_remaining - projected_calls) < (api_max * 0.2)

            if will_hit_limit:
                warnings.append(
                    f"Query will use ~{projected_calls} API calls. Org has {api_remaining} remaining (approaching 80% threshold)."
                )

            # If validate_only=False, try to get count estimate
            if not request.validate_only:
                # Try to get row count estimate (this is expensive, skip for now)
                pass

            return SOQLValidationResponse(
                valid=True,
                warnings=warnings,
                estimated_rows=estimated_rows,
                api_calls_remaining=api_remaining,
                projected_api_calls=projected_calls,
                will_hit_limit=will_hit_limit,
            )

        except Exception as e:
            return SOQLValidationResponse(
                valid=False,
                warnings=[f"Validation error: {str(e)}"],
                api_calls_remaining=0,
                projected_api_calls=0,
            )


@router.get("/quota")
async def get_quota_status(api_ctx=Depends(require_api_key)):
    """
    Get quota status for current user.
    """
    from ...middleware.quota_checker import get_quota
    from ...db import get_session

    account_id = api_ctx.account_id
    if not account_id:
        raise HTTPException(401, "Missing account_id")

    async for db in get_session():
        try:
            quota = await get_quota(str(account_id), db)
            limits = quota.get_limits()
            runs_limit = limits.get("daily_runs")
            if runs_limit is None:
                runs_limit = limits.get("monthly_rows")
            return {
                "ok": True,
                "data": {
                    "tier": quota.tier,
                    "runs_used": quota.daily_runs_used,
                    "runs_limit": runs_limit,
                    "rows_used": quota.daily_rows_used,
                    "reset_at": quota.reset_at.isoformat() if quota.reset_at else None,
                },
            }
        except Exception as e:
            raise HTTPException(500, f"Failed to get quota: {e}")


@router.get("/templates")
async def get_query_templates():
    """
    Return pre-built SOQL templates for common use cases.
    """
    templates = [
        {
            "id": "pipeline_by_stage",
            "name": "Pipeline by Stage",
            "description": "See open deals grouped by stage with totals",
            "soql": "SELECT StageName, COUNT(Id) DealCount, SUM(Amount) TotalValue FROM Opportunity WHERE IsClosed = false GROUP BY StageName ORDER BY SUM(Amount) DESC",
            "icon": "📊",
            "category": "Sales",
        },
        {
            "id": "meetings_this_week",
            "name": "Meetings This Week",
            "description": "This week's calendar events",
            "soql": "SELECT Subject, StartDateTime, WhoId, WhatId FROM Event WHERE StartDateTime >= THIS_WEEK ORDER BY StartDateTime",
            "icon": "📅",
            "category": "Activity",
        },
        {
            "id": "renewals_90d",
            "name": "Renewals in 90 Days",
            "description": "Upcoming renewal opportunities",
            "soql": "SELECT Name, AccountId, Amount, CloseDate, OwnerId FROM Opportunity WHERE CloseDate <= NEXT_N_DAYS:90 AND Type = 'Renewal' ORDER BY CloseDate",
            "icon": "🔄",
            "category": "Sales",
        },
        {
            "id": "new_leads_today",
            "name": "Today's New Leads",
            "description": "Leads created today",
            "soql": "SELECT Name, Company, Email, LeadSource, OwnerId FROM Lead WHERE CreatedDate = TODAY ORDER BY CreatedDate DESC",
            "icon": "✨",
            "category": "Leads",
        },
        {
            "id": "open_cases",
            "name": "Open Support Cases",
            "description": "Active cases by priority",
            "soql": "SELECT CaseNumber, Subject, Priority, Status, AccountId, OwnerId FROM Case WHERE IsClosed = false ORDER BY Priority, CreatedDate DESC LIMIT 200",
            "icon": "🎫",
            "category": "Service",
        },
    ]

    return {"ok": True, "data": {"templates": templates}}



"""Company List Builder API for Google Sheets sidebar."""

from __future__ import annotations

import asyncio
import json
import logging
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, Header, HTTPException, Security, status
from fastapi.security import HTTPAuthorizationCredentials
from typing import TYPE_CHECKING

# Lazy import google.cloud.bigquery to avoid GCP credential discovery hang at startup
if TYPE_CHECKING:
    pass

def _get_bq_config():
    """Lazily import BigQuery config to avoid startup hang."""
    from google.cloud import bigquery
    return bigquery.QueryJobConfig
from pydantic import BaseModel, Field
from sqlalchemy.ext.asyncio import AsyncSession

from ...auth_security import require_account, DEV_ACCOUNT_ID, http_bearer, api_key_hdr
from ... import settings as saas_settings
from ...db import get_session
from ..progress import get_redis
from ...services.list_builder_quota import ListBuilderQuota

# Lazy imports for GCP-heavy modules to avoid startup hang
_company_list_service_module = None
_list_builder_worker_module = None
_query_builder = None

# Redis key constants - duplicated to avoid module import
PRIMARY_KEYS = {
    "state": "listbuilder:export:{job_id}:state",
    "progress": "listbuilder:export:{job_id}:progress",
    "result": "listbuilder:export:{job_id}:result",
}
# Legacy fallback (older prefix)
LEGACY_KEYS = {
    "state": "list_builder:job:{job_id}:state",
    "progress": "list_builder:job:{job_id}:progress",
    "result": "list_builder:job:{job_id}:result",
}

def _get_query_builder():
    """Lazily create CompanyQueryBuilder to avoid GCP init at startup."""
    global _query_builder, _company_list_service_module
    if _query_builder is None:
        if _company_list_service_module is None:
            from ...services import company_list_service
            _company_list_service_module = company_list_service
        _query_builder = _company_list_service_module.CompanyQueryBuilder()
    return _query_builder

def _get_ListBuilderWorker():
    """Lazily import ListBuilderWorker to avoid GCP init at startup."""
    global _list_builder_worker_module
    if _list_builder_worker_module is None:
        from ...workers import list_builder_worker
        _list_builder_worker_module = list_builder_worker
    return _list_builder_worker_module.ListBuilderWorker

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/api/v2/foundrygraph/companies", tags=["company-list-builder"]
)


class CompanyFilters(BaseModel):
    industry: Optional[List[str]] = None
    country: Optional[List[str]] = None
    state: Optional[List[str]] = None
    employee_band: Optional[List[str]] = None
    revenue_band: Optional[List[str]] = None
    employee_min: Optional[int] = None
    employee_max: Optional[int] = None
    revenue_min: Optional[int] = None
    revenue_max: Optional[int] = None
    has_domain: Optional[bool] = None
    is_parent: Optional[bool] = None
    is_subsidiary: Optional[bool] = None


class SearchRequest(BaseModel):
    filters: CompanyFilters = Field(default_factory=CompanyFilters)
    limit: int = Field(25, ge=1, le=1000)
    offset: int = Field(0, ge=0)
    columns: Optional[List[str]] = None
    sort_by: Optional[str] = None
    sort_dir: Optional[str] = Field("desc", description="asc|desc")


class SearchResponse(BaseModel):
    rows: List[Dict[str, Any]]
    has_more: bool
    next_offset: Optional[int] = None


class ExportRequest(BaseModel):
    filters: CompanyFilters = Field(default_factory=CompanyFilters)
    limit: int = Field(1000, ge=1, le=20000)
    columns: Optional[List[str]] = None
    sort_by: Optional[str] = None
    sort_dir: Optional[str] = Field("desc", description="asc|desc")


class ExportStartResponse(BaseModel):
    job_id: str
    status: str
    allowed_rows: int


class ExportStatusResponse(BaseModel):
    job_id: str
    state: str
    progress: int
    result: Optional[Dict[str, Any]] = None


class FieldOptionsResponse(BaseModel):
    filters: Dict[str, List[str]]
    column_presets: List[Dict[str, Any]]
    available_columns: List[str]


class QuotaResponse(BaseModel):
    tier: str
    daily_limit: int
    daily_used: int
    daily_remaining: int
    monthly_limit: int
    monthly_used: int
    monthly_remaining: int
    preview_limit: int
    max_columns: int


async def _run_bigquery(query: str, params: List[Any], max_results: int) -> List[Any]:
    """Run a BigQuery query in a thread to avoid blocking the event loop."""
    def _run():
        QueryJobConfig = _get_bq_config()
        qb = _get_query_builder()
        job = qb.client.query(
            query, job_config=QueryJobConfig(query_parameters=params)
        )
        return list(job.result(max_results=max_results))

    return await asyncio.to_thread(_run)


async def _require_account_dev_fallback(
    bearer: Optional[HTTPAuthorizationCredentials] = Security(http_bearer),
    api_key: Optional[str] = Security(api_key_hdr),
    db: AsyncSession = Depends(get_session),
) -> str:
    """
    Use standard account auth; in development, fall back to DEV_ACCOUNT_ID if auth fails.
    """
    try:
        return await require_account(bearer=bearer, api_key=api_key, db=db)
    except HTTPException:
        if saas_settings.ENV.lower() in ("dev", "development"):
            logger.warning("Dev fallback to DEV_ACCOUNT_ID for company list builder endpoint")
            return DEV_ACCOUNT_ID
        raise


async def _get_redis_safe(required: bool = False):
    try:
        return await get_redis()
    except Exception as exc:  # noqa: BLE001
        logger.warning("Redis unavailable for list builder: %s", exc)
        if required:
            raise HTTPException(status_code=503, detail="Cache unavailable")
        return None


@router.get("/fields", response_model=FieldOptionsResponse)
async def get_field_options(
    account_id: str = Depends(_require_account_dev_fallback),
) -> FieldOptionsResponse:
    redis = await _get_redis_safe()
    cache_key = "listbuilder:fields:v2"
    if redis:
        cached = await redis.get(cache_key)
        if cached:
            try:
                payload = json.loads(cached)
                return FieldOptionsResponse(**payload)
            except Exception:
                pass

    qb = _get_query_builder()
    options = await qb.fetch_filter_options()
    payload = {
        "filters": options,
        "column_presets": qb.column_presets(),
        "available_columns": qb.available_columns(),
    }

    # Only cache if we have at least one non-empty option set to avoid caching empty defaults
    has_values = any(options.get(key) for key in ("industries", "countries", "states", "revenue_bands", "employee_bands"))
    if redis and has_values:
        try:
            await redis.set(cache_key, json.dumps(payload), ex=86400)
        except Exception:
            logger.warning("Failed to cache list builder fields")

    return FieldOptionsResponse(**payload)


@router.get("/quota", response_model=QuotaResponse)
async def get_quota(
    account_id: str = Depends(_require_account_dev_fallback),
    db: AsyncSession = Depends(get_session),
) -> QuotaResponse:
    redis = await _get_redis_safe()
    quota_service = ListBuilderQuota(db=db, redis=redis)
    tier_name, tier_config, quota = await quota_service.get_limits(account_id)

    daily_used = int(quota.list_builder_daily_exports or 0)
    monthly_used = int(quota.list_builder_monthly_exports or 0)

    return QuotaResponse(
        tier=tier_name,
        daily_limit=tier_config.list_builder_daily_rows,
        daily_used=daily_used,
        daily_remaining=max(0, tier_config.list_builder_daily_rows - daily_used),
        monthly_limit=tier_config.list_builder_monthly_rows,
        monthly_used=monthly_used,
        monthly_remaining=max(
            0, tier_config.list_builder_monthly_rows - monthly_used
        ),
        preview_limit=tier_config.list_builder_preview_rows,
        max_columns=tier_config.list_builder_max_columns,
    )


@router.post("/search", response_model=SearchResponse)
async def search_companies(
    request: SearchRequest,
    account_id: str = Depends(_require_account_dev_fallback),
    db: AsyncSession = Depends(get_session),
) -> SearchResponse:
    redis = await _get_redis_safe()
    quota_service = ListBuilderQuota(db=db, redis=redis)
    _, tier_config, _ = await quota_service.get_limits(account_id)

    limit = min(request.limit, tier_config.list_builder_preview_rows)
    filters = request.filters.dict(exclude_none=True)

    qb = _get_query_builder()
    query, params, selected_columns = qb.build_query(
        filters=filters,
        columns=request.columns or qb.DEFAULT_COLUMNS,
        limit=limit,
        offset=request.offset,
        sort_by=request.sort_by,
        sort_dir=request.sort_dir,
        include_extra_row=True,
        max_columns=tier_config.list_builder_max_columns,
    )

    rows = await _run_bigquery(query, params, max_results=limit + 1)
    has_more = len(rows) > limit
    trimmed = rows[:limit]
    formatted = qb.format_rows(trimmed, selected_columns)

    return SearchResponse(
        rows=formatted,
        has_more=has_more,
        next_offset=request.offset + limit if has_more else None,
    )


@router.post("/export", response_model=ExportStartResponse)
async def start_export(
    request: ExportRequest,
    account_id: str = Depends(_require_account_dev_fallback),
    db: AsyncSession = Depends(get_session),
    idempotency_key: Optional[str] = Header(default=None, alias="Idempotency-Key"),
) -> ExportStartResponse:
    redis = await _get_redis_safe()
    quota_service = ListBuilderQuota(db=db, redis=redis)
    limits = await quota_service.check_can_export(account_id, request.limit)

    if not limits["allowed"]:
        tier = limits.get("tier", "free")
        raise HTTPException(
            status_code=status.HTTP_429_TOO_MANY_REQUESTS,
            detail={
                "error": "quota_exceeded",
                "message": f"Export quota exceeded. {limits['daily_remaining']} rows remaining today, {limits['monthly_remaining']} this month.",
                "daily_remaining": limits["daily_remaining"],
                "monthly_remaining": limits["monthly_remaining"],
                "tier": tier,
                "upgrade_hint": "Upgrade to Scale for 10,000 rows/month" if tier != "scale" else None,
            },
        )

    active_jobs = await quota_service.get_active_job_count(account_id)
    if active_jobs >= 3:
        raise HTTPException(
            status_code=status.HTTP_429_TOO_MANY_REQUESTS,
            detail={"error": "too_many_jobs", "message": "Too many exports in flight"},
        )

    _, tier_config, _ = await quota_service.get_limits(account_id)
    filters = request.filters.dict(exclude_none=True)

    # Normalize columns via builder (without running the query)
    qb = _get_query_builder()
    _query, _params, selected_columns = qb.build_query(
        filters=filters,
        columns=request.columns or qb.DEFAULT_COLUMNS,
        limit=limits["allowed_rows"],
        offset=0,
        sort_by=request.sort_by,
        sort_dir=request.sort_dir,
        include_extra_row=False,
        max_columns=tier_config.list_builder_max_columns,
    )

    ListBuilderWorker = _get_ListBuilderWorker()
    worker = ListBuilderWorker(redis=redis)
    job_id, started = await worker.start_export_job(
        account_id=account_id,
        filters=filters,
        columns=selected_columns,
        limit=limits["allowed_rows"],
        sort_by=request.sort_by,
        sort_dir=request.sort_dir,
        idempotency_key=idempotency_key or "",
    )

    status_label = "queued" if started else "duplicate"
    return ExportStartResponse(
        job_id=job_id, status=status_label, allowed_rows=limits["allowed_rows"]
    )


@router.get("/export/{job_id}", response_model=ExportStatusResponse)
async def poll_export(
    job_id: str,
    account_id: str = Depends(require_account),
) -> ExportStatusResponse:
    redis = await _get_redis_safe(required=True)

    keyset = None
    state_raw = await redis.get(PRIMARY_KEYS["state"].format(job_id=job_id))
    if state_raw:
        keyset = PRIMARY_KEYS
    else:
        legacy_state = await redis.get(LEGACY_KEYS["state"].format(job_id=job_id))
        if legacy_state:
            keyset = LEGACY_KEYS
            state_raw = legacy_state
    if not state_raw or not keyset:
        raise HTTPException(status_code=404, detail="Job not found")

    state = (
        state_raw.decode("utf-8") if isinstance(state_raw, (bytes, bytearray)) else str(state_raw)
    )

    progress_raw = await redis.get(keyset["progress"].format(job_id=job_id))
    progress = 0
    if progress_raw:
        try:
            progress = int(progress_raw.decode("utf-8") if isinstance(progress_raw, (bytes, bytearray)) else progress_raw)
        except Exception:
            progress = 0

    result_raw = await redis.get(keyset["result"].format(job_id=job_id))
    result: Optional[Dict[str, Any]] = None
    if result_raw:
        try:
            parsed = result_raw.decode("utf-8") if isinstance(result_raw, (bytes, bytearray)) else result_raw
            result = json.loads(parsed)
        except Exception:
            result = {"raw": str(result_raw)}

    return ExportStatusResponse(
        job_id=job_id, state=state, progress=progress, result=result
    )



#!/usr/bin/env python
"""
Test runner script to check GUI-related tests after async refactor.
Run from project root: python test_runner.py
"""
import subprocess
import sys
import os
from pathlib import Path

def run_test(test_path, description):
    """Run a specific test and report results."""
    print(f"\n{'='*60}")
    print(f"Running: {description}")
    print(f"Test: {test_path}")
    print('='*60)

    result = subprocess.run(
        [sys.executable, "-m", "pytest", "-xvs", test_path],
        capture_output=True,
        text=True
    )

    if result.returncode == 0:
        print("✅ PASSED")
    else:
        print("❌ FAILED")
        print("\nOutput:")
        print(result.stdout)
        if result.stderr:
            print("\nErrors:")
            print(result.stderr)

    return result.returncode == 0

def check_imports():
    """Check if all required modules can be imported."""
    print("\n" + "="*60)
    print("Checking imports...")
    print("="*60)

    modules_to_check = [
        ("foundrymatch.desktop.gui", "GUI Module"),
        ("foundrymatch.desktop.engine_bridge", "Engine Bridge"),
        ("foundrymatch.desktop.models", "Models"),
        ("foundrymatch.core", "Core Module"),
        ("foundrymatch.core.engine", "Core Engine"),
    ]

    all_good = True
    for module_name, description in modules_to_check:
        try:
            __import__(module_name)
            print(f"✅ {description} ({module_name})")
        except ImportError as e:
            print(f"❌ {description} ({module_name}): {e}")
            all_good = False

    return all_good

def main():
    """Run all GUI-related tests."""
    # Change to project root
    project_root = Path(__file__).parent
    os.chdir(project_root)

    print(f"Running tests from: {project_root}")

    # First check imports
    imports_ok = check_imports()
    if not imports_ok:
        print("\n⚠️  Import issues detected. Tests may fail.")

    # Run specific tests
    tests = [
        ("tests/test_gui_bridge.py", "GUI Bridge Tests"),
        ("tests/test_engine.py::test_cached_score_roundtrip", "Engine Cache Test"),
        ("tests/test_engine.py::test_determine_optimal_blocking_strategy_successful_selection_dedupe", "Blocking Strategy Test"),
    ]

    passed = 0
    failed = 0

    for test_path, description in tests:
        if run_test(test_path, description):
            passed += 1
        else:
            failed += 1

    # Summary
    print("\n" + "="*60)
    print("SUMMARY")
    print("="*60)
    print(f"Total tests: {len(tests)}")
    print(f"Passed: {passed}")
    print(f"Failed: {failed}")

    if failed == 0:
        print("\n✅ All tests passed!")
    else:
        print(f"\n❌ {failed} tests failed.")

    return 0 if failed == 0 else 1

if __name__ == "__main__":
    sys.exit(main())



"""Forecast hierarchy drift check."""

from __future__ import annotations

from typing import Any, Dict, List, Tuple

from .base import BaseScoutCheck, PreflightResult
from ..crm_provider import Filter, FilterExpr, FilterOp, OrGroup
from ..models import EvidenceRow, RecommendedAction


class ForecastHierarchyDriftCheck(BaseScoutCheck):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._role_field = str(self.params.get("role_field") or "ForecastingRoleId")
        self._role_id = (self.params.get("role_id") or "").strip()

    def preflight(self) -> PreflightResult:
        base = super().preflight()
        if base.unsupported:
            return base
        try:
            desc = self.provider.describe("user")
            available = self._extract_field_names(desc)
            if self._role_field not in available:
                return PreflightResult(
                    ok=False,
                    unsupported=True,
                    reason=f"Field not found: {self._role_field}",
                )
        except Exception as exc:
            return PreflightResult(ok=False, unsupported=True, reason=str(exc))
        return base

    def build_count_filters(self) -> Tuple[str, List[FilterExpr]]:
        filters: List[FilterExpr] = []
        require_active = bool(self.params.get("require_active", True))
        if require_active:
            filters.append(Filter("IsActive", FilterOp.EQ, True))

        profile_ids = self.params.get("profile_ids") or []
        if profile_ids:
            filters.append(Filter("ProfileId", FilterOp.IN, profile_ids))

        if self._role_id:
            filters.append(
                OrGroup(
                    filters=[
                        Filter(self._role_field, FilterOp.IS_NULL),
                        Filter(self._role_field, FilterOp.NEQ, self._role_id),
                    ]
                )
            )
        else:
            filters.append(Filter(self._role_field, FilterOp.IS_NULL))

        return "user", filters

    def build_sample_request(self) -> Tuple[str, List[str], List[FilterExpr], str | None, str, bool]:
        fields = [
            "Id",
            "Name",
            "Email",
            "IsActive",
            "ProfileId",
            self._role_field,
        ]
        filters = self.build_count_filters()[1]
        return "user", fields, filters, "Name", "ASC", False

    def evaluate(self, rows: List[Dict[str, Any]]) -> List[EvidenceRow]:
        evidence: List[EvidenceRow] = []
        for row in rows:
            record_id = row.get("Id", "")
            if not record_id:
                continue
            role_value = row.get(self._role_field)
            reason_code = "MISSING_FORECAST_ROLE"
            reason = "Missing forecasting role"
            if self._role_id and role_value and role_value != self._role_id:
                reason_code = "FORECAST_ROLE_MISMATCH"
                reason = "Forecast role does not match expected role"
            fields = dict(row)
            if self._role_id:
                fields["expected_role_id"] = self._role_id
            evidence.append(
                EvidenceRow(
                    record_id=record_id,
                    sobject="user",
                    fields=fields,
                    reason=reason,
                    reason_code=reason_code,
                    severity=self.definition.severity_default,
                )
            )
        return evidence

    def get_recommended_actions(self) -> List[RecommendedAction]:
        if not self._role_id:
            return []
        return [
            RecommendedAction(
                action_type="sfdc.forecast.assign_user_to_role",
                label="Assign user to forecasting role",
                risk_level="low",
                details={"role_id": self._role_id},
            )
        ]

    def _extract_field_names(self, desc: Dict[str, Any]) -> set[str]:
        if "result" in desc:
            fields = desc.get("result", {}).get("fields", []) or []
            return {f.get("name") for f in fields if isinstance(f, dict) and f.get("name")}
        if "properties" in desc:
            props = desc.get("properties", []) or []
            return {p.get("name") for p in props if isinstance(p, dict) and p.get("name")}
        return set()


__all__ = ["ForecastHierarchyDriftCheck"]



from __future__ import annotations

import base64
from datetime import datetime, timedelta, timezone
import hashlib
from html import escape
import json
import logging
import os
import secrets
from typing import Any, Dict, Optional
from urllib.parse import urlparse

import jwt
from fastapi import APIRouter, Depends, HTTPException, Query, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from sqlalchemy.ext.asyncio import AsyncSession

from ... import settings
from ...api.progress import get_redis
from ...auth import get_current_account
from ...auth_security import require_account
from ...db import get_session
from ...models import User
from ...services.roles import require_admin, require_viewer
from ...services.secret_storage import get_saas_secret_backend
from ...services.zoominfo_oauth import (
    ZoomInfoOAuthError,
    build_zoominfo_auth_url,
    disconnect_zoominfo,
    exchange_code_for_tokens,
    fetch_zoominfo_identity,
    get_zoominfo_metadata,
    get_zoominfo_secret,
    get_zoominfo_status,
    refresh_if_needed,
    sanitize_zoominfo_error_text,
    store_zoominfo_tokens,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/api/v2/integrations/zoominfo", tags=["zoominfo"])

OAUTH_NONCE_TTL_SECONDS = 600
CLI_SESSION_TTL_SECONDS = 300


def _state_secret() -> str:
    secret = settings.ENCRYPTION_KEY or settings.JWT_SECRET_KEY
    if secret:
        return secret
    env = (settings.ENV or "").strip().lower()
    is_dev_like = env in {"dev", "development", "test"} or os.getenv("FM_TEST_MODE") == "1"
    if is_dev_like:
        logger.warning("ZoomInfo OAuth state secret missing; using dev fallback in %s environment", env or "unknown")
        return "dev-secret"
    raise RuntimeError("ZoomInfo OAuth state secret is not configured.")


def _pkce_verifier() -> str:
    # RFC 7636: high-entropy verifier in [43, 128] chars from unreserved set.
    return secrets.token_urlsafe(64)


def _pkce_challenge(verifier: str) -> str:
    digest = hashlib.sha256(verifier.encode("ascii")).digest()
    return base64.urlsafe_b64encode(digest).rstrip(b"=").decode("ascii")


def _encode_state(nonce: str, flow: str) -> str:
    payload = {
        "nonce": nonce,
        "flow": flow,
        "exp": int((datetime.now(timezone.utc) + timedelta(seconds=OAUTH_NONCE_TTL_SECONDS)).timestamp()),
    }
    return jwt.encode(payload, _state_secret(), algorithm="HS256")


def _decode_state(state_token: str) -> Dict[str, Any]:
    return jwt.decode(state_token, _state_secret(), algorithms=["HS256"])


async def _redis_get_json(redis, key: str) -> Optional[Dict[str, Any]]:
    raw = await redis.get(key)
    if not raw:
        return None
    if isinstance(raw, bytes):
        raw = raw.decode("utf-8", errors="ignore")
    try:
        parsed = json.loads(raw)
    except json.JSONDecodeError:
        return None
    return parsed if isinstance(parsed, dict) else None


async def _set_cli_session_status(
    redis,
    *,
    session_id: Optional[str],
    status: str,
    message: Optional[str] = None,
    workspace_id: Optional[str] = None,
    account_id: Optional[str] = None,
) -> None:
    if not session_id:
        return
    payload: Dict[str, Any] = {"status": status}
    if message:
        payload["message"] = message
    if workspace_id:
        payload["workspace_id"] = workspace_id
    if account_id:
        payload["account_id"] = account_id
    await redis.setex(
        f"zoominfo_cli_session:{session_id}",
        CLI_SESSION_TTL_SECONDS,
        json.dumps(payload),
    )


def _error_html(message: str) -> HTMLResponse:
    safe_message = escape(message or "Unexpected ZoomInfo OAuth failure.")
    html = f"""
    <html>
      <head><title>ZoomInfo OAuth Failed</title></head>
      <body style="font-family:Arial, sans-serif; background:#f6f7fb; color:#111;">
        <div style="max-width:560px;margin:10vh auto;background:#fff;padding:28px;border-radius:12px;">
          <h2 style="margin:0 0 8px;color:#b00020;">ZoomInfo OAuth Failed</h2>
          <p style="line-height:1.5;">{safe_message}</p>
          <p>You can close this window.</p>
        </div>
      </body>
    </html>
    """
    return HTMLResponse(content=html, status_code=200)


def _origin_from_url(url: str) -> str:
    parsed = urlparse(url)
    if parsed.scheme and parsed.netloc:
        return f"{parsed.scheme}://{parsed.netloc}"
    return url


def _success_html(frontend_url: str) -> HTMLResponse:
    safe_frontend_url = escape(frontend_url or "", quote=True)
    js_frontend_url = json.dumps(frontend_url or "")
    js_target_origin = json.dumps(_origin_from_url(frontend_url))
    html = f"""
    <html>
      <head><title>ZoomInfo Connected</title></head>
      <body style="font-family:Arial, sans-serif; background:#f6f7fb; color:#111;">
        <div style="max-width:560px;margin:10vh auto;background:#fff;padding:28px;border-radius:12px;">
          <h2 style="margin:0 0 8px;color:#0f8a3b;">ZoomInfo Connected</h2>
          <p style="line-height:1.5;">Your ZoomInfo account is connected. You can return to FoundryOps.</p>
          <p><a href="{safe_frontend_url}">Return to FoundryOps</a></p>
        </div>
        <script>
          const targetOrigin = {js_target_origin};
          const returnUrl = {js_frontend_url};
          try {{
            if (window.opener) {{
              window.opener.postMessage({{type: 'zoominfo_connected'}}, targetOrigin);
              setTimeout(function() {{ try {{ window.close(); }} catch (e) {{}} }}, 300);
            }} else {{
              window.location.assign(returnUrl);
            }}
          }} catch (e) {{}}
        </script>
      </body>
    </html>
    """
    return HTMLResponse(content=html, status_code=200)


def _resolve_redirect_uri(request: Request) -> str:
    if settings.ZOOMINFO_OAUTH_REDIRECT_URI:
        return settings.ZOOMINFO_OAUTH_REDIRECT_URI
    return str(request.url_for("zoominfo_oauth_callback"))


@router.get("/connect")
async def start_zoominfo_connect(
    request: Request,
    redirect: bool = Query(False, description="If true, issue a 302 redirect to ZoomInfo auth URL."),
    account_id=Depends(get_current_account),
    redis=Depends(get_redis),
    _user: User = Depends(require_viewer),
):
    if (
        not settings.ZOOMINFO_CLIENT_ID
        or not settings.ZOOMINFO_CLIENT_SECRET
        or not (settings.ZOOMINFO_OAUTH_AUTHORIZE_URL or "").strip()
        or not (settings.ZOOMINFO_OAUTH_TOKEN_URL or "").strip()
    ):
        raise HTTPException(status_code=503, detail="ZoomInfo OAuth not configured")

    nonce = secrets.token_urlsafe(32)
    try:
        verifier = _pkce_verifier()
        challenge = _pkce_challenge(verifier)
        state_token = _encode_state(nonce=nonce, flow="web")
    except RuntimeError as exc:
        raise HTTPException(status_code=503, detail=str(exc))

    await redis.setex(
        f"zoominfo_oauth_nonce:{nonce}",
        OAUTH_NONCE_TTL_SECONDS,
        json.dumps({"account_id": str(account_id), "code_verifier": verifier}),
    )
    try:
        auth_url = build_zoominfo_auth_url(
            state=state_token,
            redirect_uri=_resolve_redirect_uri(request),
            code_challenge=challenge,
        )
    except ZoomInfoOAuthError as exc:
        raise HTTPException(status_code=exc.status_code, detail=sanitize_zoominfo_error_text(str(exc)))
    if redirect:
        return RedirectResponse(auth_url)
    return {"auth_url": auth_url}


@router.get("/callback", name="zoominfo_oauth_callback")
async def zoominfo_oauth_callback(
    request: Request,
    code: Optional[str] = Query(None),
    state: Optional[str] = Query(None),
    error: Optional[str] = Query(None),
    error_description: Optional[str] = Query(None),
    db: AsyncSession = Depends(get_session),
    redis=Depends(get_redis),
):
    if not state:
        return _error_html("Missing OAuth state.")

    try:
        decoded = _decode_state(state)
        nonce = str(decoded.get("nonce") or "")
        flow = str(decoded.get("flow") or "")
    except RuntimeError:
        return _error_html("ZoomInfo OAuth state secret is not configured.")
    except jwt.ExpiredSignatureError:
        return _error_html("OAuth state expired. Please try again.")
    except jwt.InvalidTokenError:
        return _error_html("Invalid OAuth state.")

    if flow not in {"web", "cli"} or not nonce:
        return _error_html("Invalid OAuth session.")

    nonce_key = f"zoominfo_oauth_nonce:{nonce}"
    nonce_data = await _redis_get_json(redis, nonce_key)
    if not nonce_data:
        return _error_html("OAuth session expired. Please retry.")
    await redis.delete(nonce_key)

    account_id = str(nonce_data.get("account_id") or "")
    session_id = str(nonce_data.get("cli_session_id") or "") if flow == "cli" else None
    code_verifier = str(nonce_data.get("code_verifier") or "")
    if not account_id:
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="error",
            message="Missing workspace context.",
        )
        return _error_html("Missing workspace context.")

    if error:
        message = error_description or error
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="error",
            message=message,
            account_id=account_id,
        )
        return _error_html(f"ZoomInfo authorization failed: {message}")
    if not code:
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="error",
            message="Missing authorization code.",
            account_id=account_id,
        )
        return _error_html("Missing authorization code.")

    try:
        token_payload = await exchange_code_for_tokens(
            code=code,
            redirect_uri=_resolve_redirect_uri(request),
            code_verifier=code_verifier or None,
        )
        access_token = str(token_payload.get("access_token") or "")
        identity = await fetch_zoominfo_identity(access_token)
        metadata_overrides = {
            "org_id": identity.get("org_id"),
            "org_name": identity.get("org_name"),
            "user_id": identity.get("user_id"),
            "email": identity.get("email"),
            "name": identity.get("name"),
        }
        await store_zoominfo_tokens(
            db,
            account_id=account_id,
            tokens=token_payload,
            created_by="zoominfo_oauth",
            metadata_overrides=metadata_overrides,
        )
        await db.commit()
    except ZoomInfoOAuthError as exc:
        await db.rollback()
        safe_error = sanitize_zoominfo_error_text(str(exc))
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="error",
            message=safe_error,
            account_id=account_id,
        )
        return _error_html(safe_error)
    except Exception:  # pragma: no cover - defensive
        await db.rollback()
        logger.exception("Unexpected ZoomInfo OAuth callback failure")
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="error",
            message="Unexpected ZoomInfo OAuth failure.",
            account_id=account_id,
        )
        return _error_html("Unexpected ZoomInfo OAuth failure.")

    if flow == "cli":
        await _set_cli_session_status(
            redis,
            session_id=session_id,
            status="complete",
            workspace_id=account_id,
            account_id=account_id,
        )
        return _success_html(frontend_url=settings.FRONTEND_ORIGIN or "https://app.foundryops.io")

    frontend = settings.FRONTEND_ORIGIN or "https://app.foundryops.io"
    return _success_html(frontend_url=f"{frontend}/agent/quickstart?zoominfo_connected=1")


@router.get("/status")
async def zoominfo_status(
    db: AsyncSession = Depends(get_session),
    account_id=Depends(require_account),
):
    account_id_str = str(account_id)
    status_payload = await get_zoominfo_status(db, account_id_str)

    if status_payload.get("connected") and status_payload.get("needs_refresh"):
        try:
            await refresh_if_needed(db, account_id=account_id_str, force=False, created_by="system")
            await db.commit()
            status_payload = await get_zoominfo_status(db, account_id_str)
        except ZoomInfoOAuthError as exc:
            metadata = await get_zoominfo_metadata(db, account_id_str)
            safe_message = sanitize_zoominfo_error_text(str(exc))
            metadata.update(
                {
                    "last_refresh_status": "error",
                    "last_error_code": exc.code,
                    "last_error_message": safe_message,
                    "last_error_at": datetime.now(timezone.utc).isoformat(),
                }
            )
            secret_backend = get_saas_secret_backend()
            await secret_backend.set_secret(
                db=db,
                integration="zoominfo",
                key="oauth_metadata",
                value=json.dumps(metadata, separators=(",", ":"), sort_keys=True),
                workspace_id=account_id_str,
                created_by="system",
            )
            await db.commit()
            status_payload = await get_zoominfo_status(db, account_id_str)
            status_payload["status"] = "error"
            status_payload["error"] = {
                "code": exc.code,
                "message": safe_message,
            }
    return status_payload


@router.post("/refresh")
async def manual_refresh_zoominfo(
    db: AsyncSession = Depends(get_session),
    account_id=Depends(get_current_account),
    _user: User = Depends(require_admin),
):
    try:
        metadata = await refresh_if_needed(
            db,
            account_id=str(account_id),
            force=True,
            created_by="system",
        )
        await db.commit()
        return {
            "ok": True,
            "refreshed_at": metadata.get("refreshed_at") or metadata.get("updated_at"),
            "expires_at": metadata.get("expires_at"),
        }
    except ZoomInfoOAuthError as exc:
        await db.rollback()
        raise HTTPException(
            status_code=exc.status_code,
            detail={
                "code": exc.code,
                "message": sanitize_zoominfo_error_text(str(exc)),
            },
        )


@router.delete("/disconnect")
async def disconnect_zoominfo_route(
    db: AsyncSession = Depends(get_session),
    account_id=Depends(get_current_account),
    _user: User = Depends(require_admin),
):
    await disconnect_zoominfo(db, str(account_id))
    await db.commit()
    return {"ok": True, "message": "ZoomInfo disconnected"}


@router.get("/cli/authorize")
async def zoominfo_cli_authorize(
    request: Request,
    session_id: str = Query(...),
    workspace_id: Optional[str] = Query(None),
    account_id: str = Depends(require_account),
    redis=Depends(get_redis),
):
    if (
        not settings.ZOOMINFO_CLIENT_ID
        or not settings.ZOOMINFO_CLIENT_SECRET
        or not (settings.ZOOMINFO_OAUTH_AUTHORIZE_URL or "").strip()
        or not (settings.ZOOMINFO_OAUTH_TOKEN_URL or "").strip()
    ):
        raise HTTPException(status_code=503, detail="ZoomInfo OAuth not configured")
    target_workspace_id = str(workspace_id or account_id)
    # Optional explicit workspace override must still match caller auth context.
    if workspace_id and str(account_id) != target_workspace_id:
        raise HTTPException(status_code=403, detail="Workspace mismatch")

    nonce = secrets.token_urlsafe(32)
    try:
        verifier = _pkce_verifier()
        challenge = _pkce_challenge(verifier)
        state_token = _encode_state(nonce=nonce, flow="cli")
    except RuntimeError as exc:
        raise HTTPException(status_code=503, detail=str(exc))

    await redis.setex(
        f"zoominfo_oauth_nonce:{nonce}",
        OAUTH_NONCE_TTL_SECONDS,
        json.dumps(
            {
                "account_id": target_workspace_id,
                "cli_session_id": session_id,
                "code_verifier": verifier,
            }
        ),
    )
    await _set_cli_session_status(
        redis,
        session_id=session_id,
        status="pending",
        workspace_id=target_workspace_id,
        account_id=str(account_id),
    )
    try:
        auth_url = build_zoominfo_auth_url(
            state=state_token,
            redirect_uri=_resolve_redirect_uri(request),
            code_challenge=challenge,
        )
    except ZoomInfoOAuthError as exc:
        raise HTTPException(status_code=exc.status_code, detail=sanitize_zoominfo_error_text(str(exc)))
    return {"auth_url": auth_url, "workspace_id": target_workspace_id}


@router.get("/cli/poll")
async def zoominfo_cli_poll(
    session_id: str = Query(...),
    account_id: str = Depends(require_account),
    redis=Depends(get_redis),
):
    payload = await _redis_get_json(redis, f"zoominfo_cli_session:{session_id}")
    if not payload:
        return {"status": "pending"}
    payload_account_id = str(payload.get("account_id") or "")
    if payload_account_id and payload_account_id != str(account_id):
        raise HTTPException(status_code=403, detail="Workspace mismatch")
    return payload


@router.get("/cli/token")
async def zoominfo_cli_token(
    workspace_id: Optional[str] = Query(None),
    session_id: Optional[str] = Query(None),
    account_id: str = Depends(require_account),
    db: AsyncSession = Depends(get_session),
    redis=Depends(get_redis),
):
    caller_account_id = str(account_id)
    session_workspace_id = ""
    if session_id:
        payload = await _redis_get_json(redis, f"zoominfo_cli_session:{session_id}") or {}
        session_account_id = str(payload.get("account_id") or "")
        session_workspace_id = str(payload.get("workspace_id") or "")
        if session_account_id and session_account_id != caller_account_id:
            raise HTTPException(status_code=403, detail="Workspace mismatch")

    target_workspace_id = str(workspace_id or session_workspace_id or caller_account_id)
    if workspace_id and caller_account_id != str(workspace_id):
        raise HTTPException(status_code=403, detail="Workspace mismatch")
    if session_workspace_id and target_workspace_id != session_workspace_id:
        raise HTTPException(status_code=403, detail="Workspace mismatch")
    try:
        metadata = await refresh_if_needed(
            db,
            account_id=target_workspace_id,
            force=False,
            created_by="system",
        )
        await db.commit()
    except ZoomInfoOAuthError as exc:
        await db.rollback()
        raise HTTPException(
            status_code=exc.status_code,
            detail={
                "code": exc.code,
                "message": sanitize_zoominfo_error_text(str(exc)),
            },
        )
    access_token = await get_zoominfo_secret(db, target_workspace_id, "access_token")
    if not access_token:
        raise HTTPException(status_code=404, detail={"code": "zoominfo_not_connected", "message": "ZoomInfo is not connected."})
    return {
        "access_token": access_token,
        "org_id": metadata.get("org_id"),
        "org_name": metadata.get("org_name"),
        "expires_at": metadata.get("expires_at"),
        "scope": metadata.get("scope"),
    }



# Create performance test suite: tests/test_performance.py
import pytest
import pandas as pd
import numpy as np
from memory_profiler import profile
import gc

class TestPerformanceHardening:
    @pytest.fixture
    def mixed_script_data(self):
        """Generate 1M row mixed-script dataset."""
        np.random.seed(42)
        
        # Mix of scripts
        latin_names = ["Smith", "Johnson", "Williams", "Brown", "Jones"]
        cyrillic_names = ["Иванов", "Петров", "Сидоров", "Васильев", "Михайлов"]
        arabic_names = ["محمد", "أحمد", "علي", "حسن", "عبدالله"]
        cjk_names = ["王", "李", "张", "刘", "陈"]
        
        n = 1_000_000
        data = {
            "id": range(n),
            "name": np.random.choice(
                latin_names + cyrillic_names + arabic_names + cjk_names, 
                size=n
            ),
            "company": np.random.choice(
                ["ABC Corp", "XYZ Ltd", "テスト会社", "مؤسسة"], 
                size=n
            ),
            "email": [f"user{i}@example.com" for i in range(n)]
        }
        
        return pd.DataFrame(data)
    
    @profile
    def test_memory_usage(self, mixed_script_data):
        """Test memory usage stays within bounds."""
        from fmatch.core.engine import determine_optimal_blocking_strategy
        
        # Force garbage collection
        gc.collect()
        
        # Run blocking analysis
        result = determine_optimal_blocking_strategy(
            mixed_script_data.head(100_000),
            mappings=[{"source": "company", "ref": "company"}],
            mode="duplicate"
        )
        
        # Memory should not exceed 2GB for 100k rows
        assert result is not None
        
    def test_regex_performance(self):
        """Test optimized regex performance."""
        from fmatch.core.engine import _detect_script
        
        # Pre-compiled regex should be fast
        test_strings = ["Hello", "Привет", "مرحبا", "你好"] * 10000
        
        import time
        start = time.time()
        for s in test_strings:
            _detect_script(s)
        elapsed = time.time() - start
        
        # Should process 40k strings in under 1 second
        assert elapsed < 1.0


"""
ZoomInfo enrichment client.

Performs OAuth2 token-based calls to ZoomInfo and looks up companies by domain.
Tokens are cached in ~/.g_gremlin/config.json (keys: zoominfo_access_token,
zoominfo_refresh_token, zoominfo_token_expires_at).
"""

from __future__ import annotations

import json
import os
import time
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Optional

import click

from g_gremlin import dev_mode
from g_gremlin.base import AuthenticationError, BaseIntegration, IntegrationError, RateLimitError, RetryConfig, TransientError
from g_gremlin.credentials import resolve_credentials
from g_gremlin.credentials.provider import CredentialProvider, SecretNotFoundError

CONFIG_PATH = Path.home() / ".g_gremlin" / "config.json"
TOKEN_URL = "https://api.zoominfo.com/auth/token"
LOOKUP_URL = "https://api.zoominfo.com/lookup/company"
FOUNDRYOPS_DEFAULT_API_BASE = "https://api.foundryops.io"
DEFAULT_RETRY_CONFIG = RetryConfig(max_attempts=4, base_delay=1.0, max_delay=12.0)


class ZoomInfoError(RuntimeError):
    """Raised when ZoomInfo operations fail."""


class ZoomInfoAuthError(ZoomInfoError):
    """Raised when access tokens are invalid/expired."""


def _load_config() -> Dict[str, Any]:
    if not CONFIG_PATH.exists():
        return {}
    try:
        return json.loads(CONFIG_PATH.read_text(encoding="utf-8"))
    except Exception:
        return {}


def _save_config(config: Dict[str, Any]) -> None:
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    CONFIG_PATH.write_text(json.dumps(config, indent=2), encoding="utf-8")


def _parse_expiry(value: Any) -> Optional[float]:
    if value is None:
        return None
    if isinstance(value, (int, float)):
        return float(value)
    raw = str(value).strip()
    if not raw:
        return None
    try:
        return float(raw)
    except ValueError:
        pass
    normalized = raw.replace("Z", "+00:00")
    try:
        parsed = datetime.fromisoformat(normalized)
    except ValueError:
        return None
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)
    return parsed.timestamp()


def _resolve_foundryops_api_base(explicit: Optional[str] = None) -> str:
    return (
        explicit
        or os.getenv("FOUNDRYOPS_API_URL")
        or os.getenv("FOUNDRYMATCH_API_URL")
        or FOUNDRYOPS_DEFAULT_API_BASE
    ).rstrip("/")


def _resolve_foundryops_auth_headers() -> Dict[str, str]:
    auth_token = os.getenv("FOUNDRYMATCH_AUTH_TOKEN")
    if auth_token:
        return {"Authorization": f"Bearer {auth_token}"}

    api_key = os.getenv("FOUNDRYMATCH_API_KEY")
    if api_key:
        return {"X-API-Key": api_key}

    try:
        from g_gremlin.credentials import license as license_mod

        license_token = license_mod.get_license_token()
    except Exception:
        license_token = None
    if license_token:
        return {"Authorization": f"Bearer {license_token}"}

    if dev_mode.is_dev_mode_enabled():
        return {"Authorization": "Bearer dev-token"}
    return {}


class ZoomInfoClient(BaseIntegration):
    SERVICE_NAME = "zoominfo"

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        credential_provider: CredentialProvider | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(credential_provider=credential_provider, default_retry_config=retry_config or DEFAULT_RETRY_CONFIG)
        self.config = _load_config()
        creds = self._resolve_creds(client_id, client_secret, credential_provider)
        self._auth_mode = str(
            creds.get("auth_mode")
            or self.config.get("zoominfo_auth_mode")
            or self.config.get("auth_mode")
            or "local"
        ).strip().lower()
        self._client_id = str(creds.get("client_id") or "").strip()
        self._client_secret = str(creds.get("client_secret") or "").strip()
        self.config["zoominfo_auth_mode"] = self._auth_mode

    def _resolve_creds(
        self,
        client_id: Optional[str],
        client_secret: Optional[str],
        credential_provider: CredentialProvider | None,
    ) -> Dict[str, Any]:
        if client_id and client_secret:
            return {
                "client_id": client_id,
                "client_secret": client_secret,
                "auth_mode": "local",
            }

        auth_mode = str(self.config.get("zoominfo_auth_mode") or self.config.get("auth_mode") or "").strip().lower()
        if auth_mode == "server" and self.config.get("zoominfo_access_token"):
            return {
                "auth_mode": "server",
                "access_token": self.config.get("zoominfo_access_token"),
                "token_expires_at": self.config.get("zoominfo_token_expires_at"),
            }

        try:
            creds = resolve_credentials(
                "zoominfo",
                required_keys=["client_id", "client_secret"],
                optional_keys=["access_token", "refresh_token", "token_expires_at"],
                credential_provider=credential_provider,
            )
            # Merge into config for caching
            self.config.setdefault("zoominfo_access_token", creds.get("access_token"))
            self.config.setdefault("zoominfo_refresh_token", creds.get("refresh_token"))
            if creds.get("token_expires_at"):
                self.config.setdefault("zoominfo_token_expires_at", creds.get("token_expires_at"))
            return {
                "client_id": creds["client_id"],
                "client_secret": creds["client_secret"],
                "auth_mode": "local",
            }
        except SecretNotFoundError:
            # Fallback to legacy config file fields
            cid = (self.config.get("zoominfo_client_id") or "").strip()
            csecret = (self.config.get("zoominfo_client_secret") or "").strip()
            if cid and csecret:
                return {
                    "client_id": cid,
                    "client_secret": csecret,
                    "auth_mode": "local",
                }
            if self.config.get("zoominfo_access_token"):
                return {
                    "auth_mode": "server",
                    "access_token": self.config.get("zoominfo_access_token"),
                    "token_expires_at": self.config.get("zoominfo_token_expires_at"),
                }
            raise click.ClickException(
                "ZoomInfo is not configured. "
                "Run 'g-gremlin zoominfo oauth connect' to authenticate."
            )

    # ------------------------------------------------------------------ #
    # Token handling
    # ------------------------------------------------------------------ #
    def _token_valid(self) -> bool:
        expires_at = self.config.get("zoominfo_token_expires_at")
        access_token = self.config.get("zoominfo_access_token")
        if not access_token or not expires_at:
            return False
        expires_ts = _parse_expiry(expires_at)
        if expires_ts is None:
            return False
        return time.time() < expires_ts - 30

    def _store_tokens(self, tokens: Dict[str, Any]) -> None:
        access_token = tokens.get("access_token")
        if not access_token:
            return

        refresh_token = tokens.get("refresh_token") or self.config.get("zoominfo_refresh_token")
        expires_ts = _parse_expiry(tokens.get("expires_at"))
        if expires_ts is None:
            try:
                expires_in = float(tokens.get("expires_in", 3600))
            except Exception:
                expires_in = 3600.0
            expires_ts = time.time() + expires_in

        self.config["zoominfo_access_token"] = access_token
        if self._auth_mode == "server":
            self.config.pop("zoominfo_refresh_token", None)
            self.config.pop("zoominfo_client_id", None)
            self.config.pop("zoominfo_client_secret", None)
        else:
            self.config["zoominfo_refresh_token"] = refresh_token
        self.config["zoominfo_token_expires_at"] = expires_ts
        self.config["zoominfo_auth_mode"] = self._auth_mode
        _save_config(self.config)

    def _server_refresh_token(self) -> str:
        message = "Could not refresh token. Run 'g-gremlin zoominfo oauth connect' to re-authenticate."
        headers = _resolve_foundryops_auth_headers()
        if not headers:
            raise ZoomInfoAuthError(message)

        try:
            import requests
        except Exception as exc:
            raise ZoomInfoAuthError(message) from exc

        api_base = _resolve_foundryops_api_base(
            str(
                self.config.get("zoominfo_foundryops_api_base")
                or self.config.get("foundryops_api_base")
                or self.config.get("api_base")
                or ""
            )
        )
        params: Dict[str, str] = {}
        workspace_id = str(self.config.get("zoominfo_workspace_id") or self.config.get("workspace_id") or "").strip()
        if workspace_id:
            params["workspace_id"] = workspace_id

        try:
            resp = requests.get(
                f"{api_base}/api/v2/integrations/zoominfo/cli/token",
                params=params or None,
                headers=headers,
                timeout=20,
            )
        except Exception as exc:
            raise ZoomInfoAuthError(message) from exc

        if resp.status_code in (401, 403):
            raise ZoomInfoAuthError(message)
        if resp.status_code != 200:
            raise ZoomInfoAuthError(message)

        try:
            payload = resp.json()
        except Exception as exc:
            raise ZoomInfoAuthError(message) from exc

        access_token = str(payload.get("access_token") or "").strip()
        if not access_token:
            raise ZoomInfoAuthError(message)

        self._store_tokens(
            {
                "access_token": access_token,
                "expires_at": payload.get("expires_at"),
                "org_id": payload.get("org_id"),
                "org_name": payload.get("org_name"),
            }
        )
        return access_token

    def _request_tokens(self, refresh_token: Optional[str]) -> Dict[str, Any]:
        if self._auth_mode == "server":
            return {"access_token": self._server_refresh_token()}

        if not self._client_id or not self._client_secret:
            raise ZoomInfoAuthError("ZoomInfo client credentials missing. Reconnect with OAuth.")

        data: Dict[str, Any]
        if refresh_token:
            data = {"grant_type": "refresh_token", "refresh_token": refresh_token}
        else:
            data = {"grant_type": "client_credentials"}

        try:
            resp = self._post(
                TOKEN_URL,
                auth=(self._client_id, self._client_secret),
                json=data,
                allow_statuses=(400,),
                timeout=20,
            )
        except AuthenticationError as exc:
            raise ZoomInfoAuthError("ZoomInfo credentials rejected.") from exc
        except (RateLimitError, IntegrationError, TransientError) as exc:
            raise ZoomInfoError(str(exc)) from exc

        try:
            return resp.json()
        except Exception as exc:  # pragma: no cover - defensive
            raise ZoomInfoError(f"Invalid token response: {exc}") from exc

    def _get_access_token(self) -> str:
        if self._token_valid():
            return self.config["zoominfo_access_token"]

        if self._auth_mode == "server":
            return self._server_refresh_token()

        refresh_token = self.config.get("zoominfo_refresh_token")
        tokens = self._request_tokens(refresh_token if refresh_token else None)
        self._store_tokens(tokens)
        access_token = tokens.get("access_token")
        if not access_token:
            raise ZoomInfoError("ZoomInfo token response missing access_token.")
        return access_token

    # ------------------------------------------------------------------ #
    # Lookup
    # ------------------------------------------------------------------ #
    def _lookup_company(self, token: str, domain: str) -> Dict[str, Any]:
        headers = {"Authorization": f"Bearer {token}"}
        params = {"website": domain}
        try:
            resp = self._get(
                LOOKUP_URL,
                headers=headers,
                params=params,
                allow_statuses=(404,),
            )
        except RateLimitError as exc:
            raise ZoomInfoError(str(exc)) from exc
        except AuthenticationError as exc:
            raise ZoomInfoAuthError(str(exc)) from exc
        except (IntegrationError, TransientError) as exc:
            raise ZoomInfoError(str(exc)) from exc

        if resp.status_code == 404:
            return {}
        try:
            return resp.json()
        except Exception as exc:  # pragma: no cover - defensive
            raise ZoomInfoError(f"Invalid ZoomInfo response: {exc}") from exc

    def enrich_company(self, domain: str, fields: Optional[list[str]] = None) -> Dict[str, Any]:
        domain_clean = (domain or "").strip()
        if not domain_clean:
            return {}

        token = self._get_access_token()

        try:
            company = self._lookup_company(token, domain_clean)
        except ZoomInfoAuthError:
            # Refresh tokens and retry once.
            if self._auth_mode == "server":
                token = self._server_refresh_token()
            else:
                tokens = self._request_tokens(self.config.get("zoominfo_refresh_token"))
                self._store_tokens(tokens)
                token = tokens.get("access_token", "")
            company = self._lookup_company(token, domain_clean)

        normalized = _extract_fields(company)
        if fields:
            return {k: normalized.get(k) for k in fields if k in normalized}
        return normalized


def _extract_fields(company: Dict[str, Any]) -> Dict[str, Any]:
    if not company:
        return {}
    return {
        "name": company.get("name") or company.get("companyName") or company.get("company_name"),
        "domain": company.get("website") or company.get("domain"),
        "revenue": company.get("revenue") or company.get("annual_revenue"),
        "employee_count": company.get("employeeCount") or company.get("employee_count"),
        "industry": company.get("industry"),
        "headquarters": company.get("headquarters") or company.get("headquartersAddress"),
    }


def enrich_company(domain: str, fields: Optional[list[str]] = None) -> Dict[str, Any]:
    """
    Enrich a company by domain using ZoomInfo.
    """
    client = ZoomInfoClient()
    try:
        return client.enrich_company(domain, fields)
    except (ZoomInfoError, ZoomInfoAuthError) as exc:
        raise click.ClickException(str(exc)) from exc


__all__ = [
    "ZoomInfoError",
    "ZoomInfoAuthError",
    "ZoomInfoClient",
    "enrich_company",
]



"""Normalize phone numbers to E.164 format with extension preservation."""

from typing import Dict, Any, Optional
import re

__transform_id__ = "normalize_phone_to_e164"
__version__ = "1.2.0"
__updated__ = "2024-09-20"

# Structured error codes
ERROR_CODES = {
    "EMPTY_INPUT": "No phone number provided",
    "NO_DIGITS": "Phone number contains no digits",
    "UNSUPPORTED_COUNTRY": "Country not in supported list",
    "INVALID_FORMAT": "Phone number format not recognized",
    "TOO_SHORT": "Phone number has too few digits",
    "TOO_LONG": "Phone number exceeds E.164 maximum length",
    "PARSE_ERROR": "Unable to parse phone number",
    "INVALID_COUNTRY_CODE": "Invalid country calling code",
}

# Extension patterns
EXT_PATTERNS = [
    r"\b(?:ext\.?|extension|x|;ext=)\s*(\d{1,6})\b",
    r"#\s*(\d{1,6})",
    r",(\d{1,6})$",  # Comma followed by digits at end
]

# Compile patterns
EXT_RE = re.compile("|".join(EXT_PATTERNS), re.IGNORECASE)

# Optional dependency
try:
    import phonenumbers

    HAS_PHONENUMBERS = True
except ImportError:
    HAS_PHONENUMBERS = False


# Country codes with their digit lengths and common formats
COUNTRY_CONFIGS = {
    "US": {"code": "1", "length": 10, "name": "United States"},
    "CA": {"code": "1", "length": 10, "name": "Canada"},
    "GB": {"code": "44", "length": 10, "name": "United Kingdom"},
    "UK": {"code": "44", "length": 10, "name": "United Kingdom"},
    "AU": {"code": "61", "length": 9, "name": "Australia"},
    "DE": {"code": "49", "length": 10, "name": "Germany"},
    "FR": {"code": "33", "length": 9, "name": "France"},
    "JP": {"code": "81", "length": 10, "name": "Japan"},
    "CN": {"code": "86", "length": 11, "name": "China"},
    "IN": {"code": "91", "length": 10, "name": "India"},
    "BR": {"code": "55", "length": 11, "name": "Brazil"},
    "MX": {"code": "52", "length": 10, "name": "Mexico"},
    "ES": {"code": "34", "length": 9, "name": "Spain"},
    "IT": {"code": "39", "length": 10, "name": "Italy"},
    "NL": {"code": "31", "length": 9, "name": "Netherlands"},
    "SE": {"code": "46", "length": 9, "name": "Sweden"},
    "NO": {"code": "47", "length": 8, "name": "Norway"},
    "DK": {"code": "45", "length": 8, "name": "Denmark"},
    "FI": {"code": "358", "length": 9, "name": "Finland"},
    "CH": {"code": "41", "length": 9, "name": "Switzerland"},
    "AT": {"code": "43", "length": 10, "name": "Austria"},
    "BE": {"code": "32", "length": 9, "name": "Belgium"},
    "SG": {"code": "65", "length": 8, "name": "Singapore"},
    "HK": {"code": "852", "length": 8, "name": "Hong Kong"},
    "KR": {"code": "82", "length": 10, "name": "South Korea"},
    "NZ": {"code": "64", "length": 9, "name": "New Zealand"},
    "IL": {"code": "972", "length": 9, "name": "Israel"},
    "AE": {"code": "971", "length": 9, "name": "UAE"},
    "ZA": {"code": "27", "length": 9, "name": "South Africa"},
}


def extract_extension(value: str) -> tuple[str, Optional[str]]:
    """Extract extension from phone number string."""
    if not value:
        return value, None

    match = EXT_RE.search(value)
    if match:
        # Find which group matched
        for i, group in enumerate(match.groups()):
            if group:
                ext = group
                # Remove extension from main number
                core = value[: match.start()] + value[match.end() :]
                return core.strip(), ext

    return value, None


def normalize_phone_to_e164(
    value: str, country: Optional[str] = "US"
) -> Dict[str, Any]:
    """
    Normalize phone number to E.164 format with extension preservation.

    E.164 format:
    - Starts with + and country code
    - No spaces, parentheses, or dashes
    - Maximum 15 digits

    Examples:
        "(415) 555-0100 x456" -> {
            "value": "+14155550100",
            "extension": "456",
            "formatted": "+1 415 555 0100 x456"
        }
        "020 7946 0958" (UK) -> "+442079460958"
        "415-555-0100" -> "+14155550100"
    """
    if not value:
        return {
            "value": "",
            "extension": None,
            "formatted": "",
            "valid": False,
            "error_code": "EMPTY_INPUT",
            "error_message": ERROR_CODES["EMPTY_INPUT"],
            "reason": "empty",  # Keep for backwards compatibility
        }

    # Extract extension
    core_number, extension = extract_extension(value)

    # Try using phonenumbers library if available
    if HAS_PHONENUMBERS:
        try:
            # Parse the core number
            region = (country or "US").upper()
            if region == "UK":
                region = "GB"
            parsed = phonenumbers.parse(core_number, region)

            if phonenumbers.is_valid_number(parsed):
                e164 = phonenumbers.format_number(
                    parsed, phonenumbers.PhoneNumberFormat.E164
                )
                intl = phonenumbers.format_number(
                    parsed, phonenumbers.PhoneNumberFormat.INTERNATIONAL
                )

                # Format with extension if present
                formatted = intl
                if extension:
                    formatted = f"{intl} x{extension}"

                return {
                    "value": e164,
                    "extension": extension,
                    "formatted": formatted,
                    "valid": True,
                    "country": phonenumbers.region_code_for_number(parsed),
                    "format": "E.164",
                }
        except Exception:
            pass  # Fall through to manual parsing

    # Manual fallback parsing
    # Clean the core number
    s = str(core_number).strip()
    has_plus = s.startswith("+")

    # Extract just the digits
    digits = re.sub(r"[^\d]", "", s)

    if not digits:
        return {
            "value": "",
            "extension": extension,
            "formatted": value,
            "valid": False,
            "error_code": "NO_DIGITS",
            "error_message": ERROR_CODES["NO_DIGITS"],
            "reason": "no_digits",  # Keep for backwards compatibility
        }

    # Get country config
    country_upper = (country or "US").upper()

    # Check if country is supported
    if country_upper not in COUNTRY_CONFIGS and country_upper != "US":
        # For explicitly unsupported countries, return error
        if country and country_upper not in ["", "UNKNOWN"]:
            return {
                "value": s,
                "extension": extension,
                "valid": False,
                "error_code": "UNSUPPORTED_COUNTRY",
                "error_message": f"{ERROR_CODES['UNSUPPORTED_COUNTRY']}: {country_upper}",
                "country_assumed": country_upper,
                "formatted": value,
            }

    config = COUNTRY_CONFIGS.get(country_upper, COUNTRY_CONFIGS["US"])

    # Handle different cases
    result = None
    detected_country = None

    # Case 1: Already has country code (with or without +)
    if has_plus or len(digits) > config["length"]:
        # Check if it starts with a known country code
        for cc_key, cc_config in COUNTRY_CONFIGS.items():
            cc = cc_config["code"]
            if digits.startswith(cc):
                remaining = digits[len(cc) :]
                # Flexible length check (allow ±1 digit for variations)
                if abs(len(remaining) - cc_config["length"]) <= 1:
                    result = f"+{digits}"
                    detected_country = cc_key
                    break

        # If no match but has +, keep as is
        if not result and has_plus:
            result = f"+{digits}"
            detected_country = "UNKNOWN"

    # Case 2: Starts with 0 (common for local format in many countries)
    elif digits.startswith("0") and country_upper != "US":
        # Remove leading 0 and add country code
        local_number = digits[1:]
        if abs(len(local_number) - config["length"]) <= 1:
            result = f"+{config['code']}{local_number}"
            detected_country = country_upper

    # Case 3: Appears to be local number
    elif abs(len(digits) - config["length"]) <= 1:
        result = f"+{config['code']}{digits}"
        detected_country = country_upper

    # Case 4: US/CA special handling (1 prefix without +)
    elif country_upper in ["US", "CA"] and digits.startswith("1") and len(digits) == 11:
        result = f"+{digits}"
        detected_country = country_upper

    # Validate the result
    if result:
        # E.164 validation
        validation_error = None
        if not result.startswith("+"):
            validation_error = "INVALID_FORMAT"
            result = None
        elif len(result) > 16:  # +1 for the plus, 15 max digits
            validation_error = "TOO_LONG"
            result = None
        elif len(result) < 8:  # Minimum E.164 length
            validation_error = "TOO_SHORT"
            result = None
        elif not re.match(r"^\+\d{7,15}$", result):
            validation_error = "INVALID_FORMAT"
            result = None

    # Build result with extension
    if result:
        response = {
            "value": result,
            "extension": extension,
            "valid": True,
            "country": detected_country,
            "country_name": COUNTRY_CONFIGS.get(detected_country, {}).get(
                "name", "Unknown"
            ),
            "format": "E.164",
        }

        # Add formatted version for display
        if detected_country == "US" or detected_country == "CA":
            # Format as +1 (415) 555-0100
            if len(result) == 12:  # +1 plus 10 digits
                formatted = f"{result[:2]} ({result[2:5]}) {result[5:8]}-{result[8:]}"
            else:
                formatted = result
        elif detected_country == "GB" or detected_country == "UK":
            # Format as +44 20 7946 0958
            if len(result) == 13:  # +44 plus 10 digits
                formatted = f"{result[:3]} {result[3:5]} {result[5:9]} {result[9:]}"
            else:
                formatted = result
        else:
            formatted = result

        # Add extension to formatted version
        if extension:
            formatted = f"{formatted} x{extension}"

        response["formatted"] = formatted
        return response
    else:
        # Determine specific error
        error_code = (
            validation_error
            if "validation_error" in locals() and validation_error
            else "INVALID_FORMAT"
        )

        # Check if it's an unsupported country
        if detected_country == "UNKNOWN" or (
            country_upper and country_upper not in COUNTRY_CONFIGS
        ):
            error_code = "UNSUPPORTED_COUNTRY"

        return {
            "value": s,  # Return original if can't normalize
            "extension": extension,
            "valid": False,
            "error_code": error_code,
            "error_message": ERROR_CODES.get(error_code, "Unknown error"),
            "reason": "cannot_normalize",  # Keep for backwards compatibility
            "original": s,
            "country_assumed": country_upper,
            "country_detected": detected_country
            if "detected_country" in locals()
            else None,
            "formatted": value,  # Return original value with extension
        }



"""Scraper API endpoints for BYO Apify web scraping."""

from __future__ import annotations

from typing import Any, Dict, List, Optional
from uuid import UUID, uuid4

from fastapi import APIRouter, Depends, Header, HTTPException
from pydantic import BaseModel, Field
from sqlalchemy.ext.asyncio import AsyncSession
import logging

from ...auth_core import require_org_context, OrgContext
from ...db import get_session
from ...services.scraper_service import ScraperService
from ...services.usage_gateway import UsageGateway
from ... import settings

log = logging.getLogger(__name__)

router = APIRouter(prefix="/scraper", tags=["Scraper"])


class ValidateRequest(BaseModel):
    rows: List[Dict[str, Any]] = Field(..., description="Rows to validate")
    actor_type: str = Field(..., description="Curated actor type")
    input_mapping: Dict[str, str] = Field(default_factory=dict)
    output_fields: List[str] = Field(default_factory=list)


class TestRequest(BaseModel):
    rows: List[Dict[str, Any]] = Field(..., description="Rows to test")
    actor_type: str = Field(..., description="Curated actor type")
    input_mapping: Dict[str, str] = Field(default_factory=dict)
    output_fields: List[str] = Field(default_factory=list)
    confirm: bool = Field(False, description="Must be true to execute test")


class RunRequest(BaseModel):
    rows: List[Dict[str, Any]] = Field(..., description="Rows to scrape")
    actor_type: str = Field(..., description="Curated actor type")
    input_mapping: Dict[str, str] = Field(default_factory=dict)
    output_fields: List[str] = Field(default_factory=list)
    run_id: Optional[str] = Field(None, description="Idempotency key")
    notify_email: Optional[str] = Field(None)
    sheet_id: Optional[str] = Field(None)
    sheet_name: Optional[str] = Field(None)


async def require_byo_access(org: OrgContext, db: AsyncSession) -> None:
    gateway = UsageGateway(db)
    capabilities = await gateway.get_capabilities(org.account_id)
    byo = capabilities.get("byo") or {}
    if not byo.get("enabled"):
        raise HTTPException(
            status_code=403,
            detail={
                "error": "TIER_REQUIRED",
                "message": "BYO scraping requires Unleashed tier or BYO add-on",
                "current_tier": capabilities.get("tier", "free"),
                "required_tier": "unleashed_or_addon",
                "upgrade_url": "/pricing",
            },
        )


@router.get("/actors")
async def list_actors(
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    await require_byo_access(org, db)
    service = ScraperService(db, org.account_id)
    actors = await service.list_actors()
    return {"ok": True, "data": {"actors": actors}}


@router.post("/validate")
async def validate_scraper(
    request_body: ValidateRequest,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    await require_byo_access(org, db)
    service = ScraperService(db, org.account_id)
    try:
        result = await service.validate_inputs(
            rows=request_body.rows,
            input_mapping=request_body.input_mapping,
            actor_type=request_body.actor_type,
        )
        return {"ok": True, "data": result.to_dict()}
    except Exception as exc:
        log.error("Scraper validation error: %s", exc, exc_info=True)
        raise HTTPException(
            status_code=500,
            detail={"error": "VALIDATION_ERROR", "message": str(exc)},
        )


@router.post("/test")
async def test_scraper(
    request_body: TestRequest,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    await require_byo_access(org, db)

    if not request_body.confirm:
        raise HTTPException(
            status_code=400,
            detail={
                "error": "CONFIRMATION_REQUIRED",
                "message": "Set confirm=true to execute live test.",
            },
        )

    service = ScraperService(db, org.account_id)
    validation = await service.validate_inputs(
        rows=request_body.rows,
        input_mapping=request_body.input_mapping,
        actor_type=request_body.actor_type,
    )
    if not validation.byo_provider_connected:
        raise HTTPException(
            status_code=400,
            detail={
                "error": "APIFY_NOT_CONNECTED",
                "message": "Apify not connected. Please add your API token.",
            },
        )
    test_rows = min(validation.rows_with_valid_inputs, len(request_body.rows), 5)

    gateway = UsageGateway(db)
    consume = await gateway.check_and_consume(
        account_id=org.account_id,
        resource="byo_rows",
        amount=test_rows,
        source="byo_scraper_test",
    )
    if not consume.allowed:
        raise HTTPException(
            status_code=429,
            detail={
                "error": "BYO_LIMIT_EXCEEDED",
                "message": consume.message or "BYO daily limit reached.",
                "limit": consume.limit,
                "used": consume.used,
                "remaining": consume.remaining,
                "reset_at": consume.reset_at.isoformat() if consume.reset_at else None,
            },
        )

    try:
        result = await service.run_test(
            rows=request_body.rows,
            input_mapping=request_body.input_mapping,
            output_fields=request_body.output_fields,
            actor_type=request_body.actor_type,
            max_rows=5,
        )
        return {"ok": True, "data": result.to_dict()}
    except Exception as exc:
        log.error("Scraper test error: %s", exc, exc_info=True)
        raise HTTPException(
            status_code=500,
            detail={"error": "SCRAPER_TEST_ERROR", "message": str(exc)},
        )


@router.post("/run")
async def run_scraper(
    request_body: RunRequest,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
    idempotency_key: Optional[str] = Header(default=None, alias="Idempotency-Key"),
) -> Dict[str, Any]:
    await require_byo_access(org, db)

    if len(request_body.rows) > settings.SCRAPER_MAX_ROWS:
        raise HTTPException(
            status_code=400,
            detail={
                "error": "ROW_LIMIT_EXCEEDED",
                "message": f"Row limit exceeded: max {settings.SCRAPER_MAX_ROWS} rows per request",
                "max_rows": settings.SCRAPER_MAX_ROWS,
                "received_rows": len(request_body.rows),
            },
        )

    run_id = (idempotency_key or request_body.run_id or "").strip()
    if not run_id:
        run_id = uuid4().hex

    service = ScraperService(db, org.account_id)
    validation = await service.validate_inputs(
        rows=request_body.rows,
        input_mapping=request_body.input_mapping,
        actor_type=request_body.actor_type,
    )
    if not validation.byo_provider_connected:
        raise HTTPException(
            status_code=400,
            detail={
                "error": "APIFY_NOT_CONNECTED",
                "message": "Apify not connected. Please add your API token.",
            },
        )

    gateway = UsageGateway(db)
    consume = await gateway.check_and_consume(
        account_id=org.account_id,
        resource="byo_rows",
        amount=validation.estimated_credits,
        source="byo_scraper_run",
        idempotency_key=f"scraper:{org.account_id}:{run_id}",
    )
    if not consume.allowed:
        raise HTTPException(
            status_code=429,
            detail={
                "error": "BYO_LIMIT_EXCEEDED",
                "message": consume.message or "BYO daily limit reached.",
                "limit": consume.limit,
                "used": consume.used,
                "remaining": consume.remaining,
                "reset_at": consume.reset_at.isoformat() if consume.reset_at else None,
            },
        )

    try:
        result = await service.create_job(
            rows=request_body.rows,
            actor_type=request_body.actor_type,
            input_mapping=request_body.input_mapping,
            output_fields=request_body.output_fields,
            run_id=run_id,
            created_by=org.email,
            notify_email=request_body.notify_email,
            sheet_id=request_body.sheet_id,
            sheet_name=request_body.sheet_name,
        )
        return {"ok": True, "data": result.to_dict()}
    except ValueError as exc:
        raise HTTPException(
            status_code=400,
            detail={"error": "VALIDATION_ERROR", "message": str(exc)},
        )
    except Exception as exc:
        log.error("Scraper job creation error: %s", exc, exc_info=True)
        raise HTTPException(
            status_code=500,
            detail={"error": "JOB_CREATION_ERROR", "message": str(exc)},
        )


@router.get("/run/{job_id}")
async def get_scraper_status(
    job_id: UUID,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    await require_byo_access(org, db)
    service = ScraperService(db, org.account_id)
    status = await service.get_job_status(job_id)
    if status is None:
        raise HTTPException(
            status_code=404,
            detail={"error": "NOT_FOUND", "message": f"Job {job_id} not found"},
        )
    return {"ok": True, "data": status.to_dict()}


@router.get("/run/{job_id}/results")
async def get_scraper_results(
    job_id: UUID,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
    cursor: int = 0,
    limit: int = 250,
) -> Dict[str, Any]:
    await require_byo_access(org, db)
    service = ScraperService(db, org.account_id)
    try:
        data = await service.get_results_chunk(job_id, cursor, limit)
    except Exception as exc:
        log.error("Scraper results fetch error: %s", exc, exc_info=True)
        raise HTTPException(
            status_code=500,
            detail={"error": "RESULTS_ERROR", "message": str(exc)},
        )
    return {"ok": True, "data": data}


@router.post("/run/{job_id}/cancel")
async def cancel_scraper_job(
    job_id: UUID,
    org: OrgContext = Depends(require_org_context),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    await require_byo_access(org, db)
    service = ScraperService(db, org.account_id)
    ok = await service.cancel_job(job_id)
    if not ok:
        raise HTTPException(
            status_code=404,
            detail={"error": "NOT_FOUND", "message": f"Job {job_id} not found"},
        )
    return {"ok": True}



from __future__ import annotations

import json
import time
import uuid
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, HTTPException, Request
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator

from .deps import require_api_key
from ...db import get_session
from ...repos.salesforce_ai_reports_repo import SalesforceAIReportsRepo
from ...services.ai_rate_limiter import OrgLimiter
from ...services.salesforce_ai_reports import (
    PlannerUnavailable,
    PlanValidationError,
    SalesforceAIReportService,
    SalesforcePlan,
    PlanPatch,
    _ensure_patch_summary,
    hash_plan_for_storage,
)
from ...services.salesforce_gateway import SalesforceGateway
from ...services.telemetry import emit
from ...settings import (
    AI_REPORTS_ENABLED,
    AI_REPORTS_MAX_CONCURRENT_GLOBAL,
    AI_REPORTS_MAX_CONCURRENT_PER_ORG,
    AI_REPORTS_MAX_FIELDS,
    AI_REPORTS_MAX_OBJECTS,
    AI_REPORTS_PREVIEW_LIMIT,
)
from ...utils.salesforce_errors import normalize_sf_error


router = APIRouter(prefix="/salesforce/ai-reports", tags=["Salesforce AI Reports"])
service = SalesforceAIReportService()
limiter = OrgLimiter(
    AI_REPORTS_MAX_CONCURRENT_PER_ORG,
    AI_REPORTS_MAX_CONCURRENT_GLOBAL,
)


class ConversationTurn(BaseModel):
    model_config = ConfigDict(extra="forbid")

    role: str
    content: str

    @field_validator("role")
    @classmethod
    def _normalize_role(cls, value: str) -> str:
        if not value:
            return "user"
        return value.lower()


class PlanRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")

    prompt: str
    conversation: List[ConversationTurn] = Field(default_factory=list)
    prior_warnings: List[str] = Field(default_factory=list)
    clarifier_answers: Dict[str, str] = Field(
        default_factory=dict, alias="clarifierAnswers"
    )


class DiscoverRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")

    prompt: str = ""


class PatchRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")

    plan: Dict[str, Any]
    patch: Dict[str, Any]
    prompt: Optional[str] = None


class PreviewRequest(BaseModel):
    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    plan: Dict[str, Any]
    tier_override: Optional[str] = Field(default=None, alias="tier")


class SaveReportRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str
    description: Optional[str] = None
    plan: Dict[str, Any]
    compiled_soql: str
    scopes: List[str] = Field(default_factory=list)
    schedule: Optional[Dict[str, Any]] = None


class ReportResponse(BaseModel):
    model_config = ConfigDict(extra="allow")

    id: str
    name: str
    description: Optional[str]
    primary_object: str
    plan: Dict[str, Any]
    compiled_soql: str
    created_at: str
    updated_at: str
    tier: Optional[str]
    plan_hash: str


async def _get_gateway(account_id: str, db_session) -> SalesforceGateway:
    return await SalesforceGateway.for_tenant(str(account_id), db_session)


@router.get("/config")
async def fetch_config():
    return {
        "enabled": AI_REPORTS_ENABLED,
        "previewLimit": AI_REPORTS_PREVIEW_LIMIT,
        "maxFields": AI_REPORTS_MAX_FIELDS,
        "maxObjects": AI_REPORTS_MAX_OBJECTS,
    }


@router.get("/schema")
async def fetch_schema_context(api_ctx=Depends(require_api_key)):
    account_id = api_ctx.account_id
    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        context = await service.build_schema_context(
            account_id=account_id,
            gateway=gateway,
            prompt="",
            db_session=db,
        )
        return {"ok": True, "data": context.to_payload()}


@router.get("/schema/peek")
async def peek_schema_field(
    object: str, field: str, limit: int = 3, api_ctx=Depends(require_api_key)
):
    if not object or not field:
        raise HTTPException(400, detail="object and field parameters are required")
    account_id = api_ctx.account_id
    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        try:
            payload = await service.peek_schema_field(
                account_id=account_id,
                gateway=gateway,
                object_name=object,
                field_name=field,
                sample_limit=limit,
            )
        except PlanValidationError as exc:
            raise HTTPException(400, detail=str(exc)) from exc
        return {"ok": True, "data": payload}


@router.post("/discover")
async def discover_context(
    body: DiscoverRequest, request: Request, api_ctx=Depends(require_api_key)
):
    account_id = api_ctx.account_id
    request_id = request.headers.get("X-Request-ID") or str(uuid.uuid4())
    start = time.perf_counter()
    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        research, schema_context = await service.discover(
            account_id=account_id,
            prompt=body.prompt,
            gateway=gateway,
            db_session=db,
            request_id=request_id,
        )
        ranked = getattr(research, "ranked_fields", None) or []
        suggested = getattr(research, "suggested_fields", None) or []
        clarifiers = getattr(research, "clarifiers", None) or []
        index_meta = getattr(research, "index_status", None)
        emit(
            "discover.ttfb_ms",
            request_id=request_id,
            org_id=account_id,
            ms=int((time.perf_counter() - start) * 1000),
            suggested=len(ranked) if ranked else len(suggested),
            clarifiers=len(clarifiers),
        )
        return {
            "ok": True,
            "discover": research.model_dump(by_alias=True),
            "schema": schema_context.to_payload(),
            "index_status": index_meta,
            "request_id": request_id,
        }


@router.post("/plan")
async def plan_report(
    body: PlanRequest, request: Request, api_ctx=Depends(require_api_key)
):
    account_id = api_ctx.account_id
    tier = getattr(api_ctx, "tier", None)
    request_id = request.headers.get("X-Request-ID") or str(uuid.uuid4())

    if body.clarifier_answers:
        for clarifier_id, choice in body.clarifier_answers.items():
            emit(
                "clarifier.picked",
                request_id=request_id,
                org_id=account_id,
                clarifier_id=clarifier_id,
                choice=choice,
            )

    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        try:
            schema_context = await service.build_schema_context(
                account_id=account_id,
                gateway=gateway,
                prompt=body.prompt,
                db_session=db,
            )
            artifacts = await service.plan_query(
                account_id=account_id,
                tier=tier,
                prompt=body.prompt,
                gateway=gateway,
                prior_warnings=body.prior_warnings,
                conversation=[turn.model_dump() for turn in body.conversation],
                schema_context=schema_context,
                clarifier_answers=body.clarifier_answers,
                db_session=db,
                request_id=request_id,
            )

            row_cap = service.tier_row_cap(tier)
            warnings = service.validate_plan(
                plan=artifacts.plan,
                schema_context=schema_context,
                row_cap=row_cap,
            )

            soql = service.compile_plan(artifacts.plan)
            plan_hash = hash_plan_for_storage(artifacts.plan)
            clarifiers_payload = [
                clarifier.model_dump(by_alias=True)
                for clarifier in artifacts.clarifiers
            ]
            alternatives_payload = [
                alt.model_dump(by_alias=True) for alt in artifacts.alternatives
            ]
            for alt in alternatives_payload:
                alt.setdefault("patchSummary", alt.get("patch", {}).get("summary"))
            discover_payload = (
                artifacts.research.model_dump(by_alias=True)
                if artifacts.research
                else None
            )

            return {
                "ok": True,
                "plan": artifacts.plan.model_dump(by_alias=True),
                "compiled_soql": soql,
                "trace": artifacts.trace,
                "clarifiers": clarifiers_payload,
                "risks": artifacts.risks,
                "alternatives": alternatives_payload,
                "discover": discover_payload,
                "warnings": warnings,
                "plan_hash": plan_hash,
                "preview_limit": AI_REPORTS_PREVIEW_LIMIT,
                "row_cap": row_cap,
                "request_id": request_id,
            }
        except PlannerUnavailable as exc:
            raise HTTPException(503, detail=str(exc)) from exc
        except PlanValidationError as exc:
            raise HTTPException(400, detail=str(exc)) from exc


@router.post("/patch")
async def patch_plan(body: PatchRequest, api_ctx=Depends(require_api_key)):
    account_id = api_ctx.account_id
    tier = getattr(api_ctx, "tier", None)

    try:
        base_plan = SalesforcePlan.model_validate(body.plan)
    except ValidationError as exc:
        raise HTTPException(400, detail=f"Invalid plan payload: {exc}") from exc

    try:
        plan_patch = PlanPatch.model_validate(body.patch)
    except ValidationError as exc:
        raise HTTPException(400, detail=f"Invalid patch payload: {exc}") from exc

    summary = _ensure_patch_summary(plan_patch).summary

    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        try:
            updated_plan = service.apply_patch(base_plan, plan_patch)
            schema_context = await service.build_schema_context(
                account_id=account_id,
                gateway=gateway,
                prompt=body.prompt
                or (updated_plan.objects[0] if updated_plan.objects else ""),
            )
            warnings = service.validate_plan(
                plan=updated_plan,
                schema_context=schema_context,
                row_cap=service.tier_row_cap(tier),
            )
        except PlanValidationError as exc:
            raise HTTPException(400, detail=str(exc)) from exc

        soql = service.compile_plan(updated_plan)
        plan_hash = hash_plan_for_storage(updated_plan)
        return {
            "ok": True,
            "plan": updated_plan.model_dump(by_alias=True),
            "compiled_soql": soql,
            "warnings": warnings,
            "patch_summary": summary,
            "plan_hash": plan_hash,
        }


@router.post("/preview")
async def preview_plan(
    body: PreviewRequest, request: Request, api_ctx=Depends(require_api_key)
):
    account_id = api_ctx.account_id
    tier = body.tier_override or getattr(api_ctx, "tier", None)
    request_id = request.headers.get("X-Request-ID") or str(uuid.uuid4())

    try:
        plan = SalesforcePlan.model_validate(body.plan)
    except ValidationError as exc:
        raise HTTPException(400, detail=f"Invalid plan payload: {exc}") from exc

    limiter.acquire(account_id)
    try:
        async for db in get_session():
            gateway = await _get_gateway(account_id, db)
            try:
                focus_prompt = plan.objects[0] if plan.objects else ""
                schema_context = await service.build_schema_context(
                    account_id=account_id,
                    gateway=gateway,
                    prompt=focus_prompt,
                    db_session=db,
                )
                warnings = service.validate_plan(
                    plan=plan,
                    schema_context=schema_context,
                    row_cap=service.tier_row_cap(tier),
                )
                preview = await service.preview_plan(
                    plan=plan,
                    gateway=gateway,
                    account_id=account_id,
                    tier=tier,
                    request_id=request_id,
                )
                return {
                    "ok": True,
                    "preview": preview,
                    "warnings": warnings,
                    "request_id": request_id,
                }
            except PlanValidationError as exc:
                raise HTTPException(400, detail=str(exc)) from exc
            except Exception as exc:
                payload = getattr(exc, "payload", None)
                if not payload and hasattr(exc, "invalid_fields"):
                    payload = {
                        "errorCode": "INSUFFICIENT_ACCESS",
                        "message": ", ".join(getattr(exc, "invalid_fields", [])),
                    }
                normalized = normalize_sf_error(
                    payload
                    or {"message": str(exc), "error": getattr(exc, "code", "UNKNOWN")}
                )
                status_code = 429 if normalized.kind == "rate_limit" else 400
                detail = normalized.model_dump()
                suggestions = service.suggest_autofix(
                    plan=plan, error_message=normalized.message
                )
                if suggestions:
                    autofix_payload = []
                    for suggestion in suggestions:
                        data = suggestion.model_dump(by_alias=True)
                        data.setdefault(
                            "patchSummary", data.get("patch", {}).get("summary")
                        )
                        autofix_payload.append(data)
                    detail["autofix"] = autofix_payload
                raise HTTPException(status_code=status_code, detail=detail) from exc
    finally:
        limiter.release(account_id)


@router.post("/reports")
async def save_report(body: SaveReportRequest, api_ctx=Depends(require_api_key)):
    account_id = api_ctx.account_id
    user_id = getattr(api_ctx, "api_key_prefix", "api")
    tier = getattr(api_ctx, "tier", None)

    try:
        plan = SalesforcePlan.model_validate(body.plan)
    except ValidationError as exc:
        raise HTTPException(400, detail=f"Invalid plan payload: {exc}") from exc

    if not body.name.strip():
        raise HTTPException(400, detail="Name is required")

    async for db in get_session():
        gateway = await _get_gateway(account_id, db)
        try:
            schema_context = await service.build_schema_context(
                account_id=account_id,
                gateway=gateway,
                prompt=plan.objects[0] if plan.objects else "",
                db_session=db,
            )
            service.validate_plan(
                plan=plan,
                schema_context=schema_context,
                row_cap=service.tier_row_cap(tier),
            )
        except PlanValidationError as exc:
            raise HTTPException(400, detail=str(exc)) from exc

        repo = SalesforceAIReportsRepo(db)
        plan_hash = hash_plan_for_storage(plan)
        record = await repo.create_report(
            account_id=account_id,
            user_id=user_id,
            name=body.name,
            description=body.description,
            primary_object=plan.objects[0] if plan.objects else "Lead",
            plan_json=plan.model_dump(by_alias=True),
            compiled_soql=body.compiled_soql,
            plan_hash=plan_hash,
            tier=tier,
            scopes=body.scopes,
            schedule=body.schedule,
        )
        return {
            "ok": True,
            "report": ReportResponse(
                id=record.id,
                name=record.name,
                description=record.description,
                primary_object=record.primary_object,
                plan=body.plan,
                compiled_soql=body.compiled_soql,
                created_at=record.created_at.isoformat(),
                updated_at=record.updated_at.isoformat(),
                tier=record.tier,
                plan_hash=plan_hash,
            ).model_dump(),
        }


@router.get("/reports")
async def list_reports(api_ctx=Depends(require_api_key)):
    account_id = api_ctx.account_id

    async for db in get_session():
        repo = SalesforceAIReportsRepo(db)
        rows = await repo.list_reports(account_id=account_id)
        out: List[Dict[str, Any]] = []
        for row in rows:
            try:
                plan_payload = json.loads(row.plan_json)
            except Exception:
                plan_payload = {}
            out.append(
                {
                    "id": row.id,
                    "name": row.name,
                    "description": row.description,
                    "primary_object": row.primary_object,
                    "plan": plan_payload,
                    "compiled_soql": row.compiled_soql,
                    "created_at": row.created_at.isoformat(),
                    "updated_at": row.updated_at.isoformat(),
                    "tier": row.tier,
                    "plan_hash": row.plan_hash,
                }
            )
        return {"ok": True, "reports": out}



import tkinter as tk
from tkinter import ttk


class IDSelectionDialog(tk.Toplevel):
    """
    Placeholder dialog for ID selection.
    This will be fully implemented later.
    """

    def __init__(self, parent, df, report):
        super().__init__(parent)
        self.title("Select ID Column")
        self.result = None  # This is important for the GUI logic

        ttk.Label(self, text="ID Selection Dialog (Placeholder)").pack(padx=20, pady=20)

        # This placeholder immediately closes, allowing high-confidence
        # detection to proceed without user interaction for now.
        self.after(100, self.destroy)



"""CLI commands for Deal Desk automation under Autopilot."""

from __future__ import annotations

import json
import os
import sys
from pathlib import Path

import click

from g_gremlin.dealdesk.config import load_dealdesk_config, DealDeskConfigError
from g_gremlin.dealdesk.planner import build_dealdesk_plan
from g_gremlin.scout.engine import ScoutEngine
from g_gremlin.scout.loader import load_scan_config

from .approvals import ApprovalTokenStore, default_approval_ttl_seconds
from .contracts import PlanV1, PlanStep
from .executor import PlanExecutor
from .policy import apply_policy_overlays, compute_policy_hash, load_policy
from .receipts import ReceiptWriter
from .slack_blocks import build_dealdesk_digest_blocks
from .slack_notify import create_client as slack_create_client
from .slack_notify import resolve_channel
from .state import resolve_state_dir
from .tools import get_default_registry


EXIT_SUCCESS = 0
EXIT_USAGE = 2
EXIT_ISSUES = 1


def _get_user_credentials() -> set[str]:
    try:
        from g_gremlin.credentials.provider import CredentialProvider

        provider = CredentialProvider()
        return set(provider.list_configured_integrations())
    except Exception:
        return set()


@click.group("dealdesk")
def dealdesk_group() -> None:
    """Deal desk automation (close verification, renewals, stamps)."""


@dealdesk_group.group("check")
def dealdesk_check_group() -> None:
    """Manage Deal Desk scout checks."""


@dealdesk_check_group.command("list")
def dealdesk_check_list() -> None:
    """List Deal Desk scout checks."""
    engine = ScoutEngine(user_credentials=_get_user_credentials())
    scouts = [s for s in engine.list_scouts() if s.name.startswith("dealdesk")]
    if not scouts:
        click.echo("No Deal Desk scouts found.")
        return
    for scout in scouts:
        click.echo(f"{scout.display_name} ({scout.name})")


@dealdesk_group.command("run")
@click.option("--since-days", default=7, type=int, show_default=True)
@click.option("--org-alias", default=None, help="Salesforce org alias.")
@click.option(
    "--config",
    "config_path",
    default=None,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Deal Desk config path (default: config/dealdesk/dealdesk.yaml).",
)
@click.option(
    "--policy",
    "policy_path",
    default=Path("autopilot_policy.yaml"),
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Autopilot policy YAML path (default: ./autopilot_policy.yaml).",
)
@click.option(
    "--output",
    "output_path",
    default=None,
    type=click.Path(dir_okay=False, path_type=Path),
    help="Write plan JSON to this path.",
)
@click.option("--slack", "send_slack", is_flag=True, help="Post Deal Desk digest to Slack.")
@click.option(
    "--artifact-base-url",
    "artifact_base_url",
    default=None,
    help="Base URL for artifact links.",
)
def dealdesk_run(
    since_days: int,
    org_alias: str | None,
    config_path: Path | None,
    policy_path: Path,
    output_path: Path | None,
    send_slack: bool,
    artifact_base_url: str | None,
) -> None:
    """Run Deal Desk close verification and generate a plan."""
    try:
        config = load_dealdesk_config(config_path or None)
    except DealDeskConfigError as exc:
        click.echo(f"Error loading Deal Desk config: {exc}", err=True)
        sys.exit(EXIT_USAGE)

    try:
        policy = load_policy(policy_path)
    except Exception as exc:
        click.echo(f"Error loading policy: {exc}", err=True)
        sys.exit(EXIT_USAGE)

    state_dir = resolve_state_dir(policy)
    scan_cfg = load_scan_config()
    crm = (scan_cfg.crm or "salesforce").lower()

    engine = ScoutEngine(user_credentials=_get_user_credentials())
    params = {
        "since_days": since_days,
        "config_path": str(config_path) if config_path else "",
    }
    scan_result = engine.run_scan(
        ["dealdesk_close_verification"],
        top_n=scan_cfg.top_n,
        per_scout_params={"dealdesk_close_verification": params},
        org_alias=org_alias or policy.org_alias or scan_cfg.org_alias,
        crm=crm,
        hubspot_profile=scan_cfg.hubspot_profile,
        dynamics_profile=scan_cfg.dynamics_profile,
    )

    run_dir = state_dir / "receipts" / scan_result.run_id
    run_dir.mkdir(parents=True, exist_ok=True)
    scan_path = run_dir / "scan.json"
    scan_path.write_text(
        json.dumps(scan_result.to_dict(include_fields=True, include_reason=True), indent=2),
        encoding="utf-8",
    )

    registry = get_default_registry()
    plan = build_dealdesk_plan(
        scan_result,
        registry,
        config,
        state_dir=state_dir,
        artifact_base_url=artifact_base_url or os.getenv("G_GREMLIN_ARTIFACT_BASE_URL"),
        config_path=config_path,
    )

    plan.policy_hash = compute_policy_hash(apply_policy_overlays(policy, state_dir))
    plan_path = run_dir / "plan.json"
    plan_path.write_text(json.dumps(plan.to_dict(), indent=2), encoding="utf-8")

    if output_path:
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_text(json.dumps(plan.to_dict(), indent=2), encoding="utf-8")

    approval_token = None
    if plan.steps:
        store = ApprovalTokenStore(state_dir / "db" / "approval_tokens.sqlite")
        approval_token = store.create_token(
            plan_hash=plan.plan_hash or "",
            policy_hash=plan.policy_hash or "",
            run_id=plan.run_id,
            ttl_seconds=default_approval_ttl_seconds(),
        ).token
        pending_dir = state_dir / "pending" / plan.run_id
        pending_dir.mkdir(parents=True, exist_ok=True)
        (pending_dir / "plan.json").write_text(json.dumps(plan.to_dict(), indent=2), encoding="utf-8")

    if send_slack:
        bot_token = os.getenv("G_GREMLIN_SLACK_BOT_TOKEN")
        if not bot_token:
            click.echo("Slack bot token missing (G_GREMLIN_SLACK_BOT_TOKEN).", err=True)
        else:
            client = slack_create_client(bot_token)
            channel = resolve_channel(policy.defaults)
            if not channel:
                click.echo("Slack channel not configured.", err=True)
            else:
                index_path = (
                    state_dir / "artifacts" / scan_result.run_id / "close_packets" / "close_packets_index.json"
                )
                blocks = build_dealdesk_digest_blocks(
                    scan_result,
                    plan=plan,
                    approve_token=approval_token,
                    index_path=index_path,
                )
                text = f"Deal Desk digest for run {scan_result.run_id}"
                client.chat_postMessage(channel=channel, text=text, **blocks)

    click.echo(json.dumps(plan.to_dict(), indent=2))


@dealdesk_group.command("apply")
@click.argument(
    "plan_path",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option(
    "--policy",
    "policy_path",
    default=Path("autopilot_policy.yaml"),
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Policy YAML path (default: ./autopilot_policy.yaml).",
)
@click.option("--approve-token", default=None, help="Approval token.")
@click.option("--dry-run", is_flag=True, help="Dry run execution.")
def dealdesk_apply(
    plan_path: Path,
    policy_path: Path,
    approve_token: str | None,
    dry_run: bool,
) -> None:
    """Apply a Deal Desk plan and write a receipt."""
    try:
        plan_payload = json.loads(plan_path.read_text(encoding="utf-8"))
        plan = PlanV1.from_dict(plan_payload)
    except Exception as exc:
        click.echo(f"Error loading plan: {exc}", err=True)
        sys.exit(EXIT_USAGE)

    try:
        policy = load_policy(policy_path)
    except Exception as exc:
        click.echo(f"Error loading policy: {exc}", err=True)
        sys.exit(EXIT_USAGE)

    state_dir = resolve_state_dir(policy)
    policy = apply_policy_overlays(policy, state_dir)
    executor = PlanExecutor(policy=policy, registry=get_default_registry(), state_dir=state_dir)
    receipt = executor.execute(plan, approve_token=approve_token, dry_run=dry_run)
    writer = ReceiptWriter(state_dir)
    receipt_dir = writer.write(receipt, plan)
    click.echo(json.dumps({"status": receipt.outcome, "receipt_dir": str(receipt_dir)}, indent=2))


@dealdesk_group.group("renewal")
def dealdesk_renewal_group() -> None:
    """Renewal helper commands."""


@dealdesk_renewal_group.command("create")
@click.option("--opp-id", required=True, help="Source Opportunity ID.")
@click.option("--org-alias", default=None, help="Salesforce org alias.")
@click.option(
    "--config",
    "config_path",
    default=None,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Deal Desk config path (default: config/dealdesk/dealdesk.yaml).",
)
def dealdesk_renewal_create(opp_id: str, org_alias: str | None, config_path: Path | None) -> None:
    """Create a renewal shell for a single opportunity."""
    from g_gremlin.dealdesk.handlers import handle_dealdesk_renewal_create_shell

    registry = get_default_registry()
    tool_def = registry.get_tool("dealdesk.renewal.create_shell")
    step = PlanStep(
        step_id="manual-renewal",
        tool="dealdesk.renewal.create_shell",
        args={
            "record_ids": [opp_id],
            "org_alias": org_alias,
            "config_path": str(config_path) if config_path else "",
        },
        risk_level=tool_def.risk_level if tool_def else 2,
    )
    result = handle_dealdesk_renewal_create_shell(step, tool_def)
    click.echo(json.dumps(result, indent=2))


@dealdesk_group.group("config")
def dealdesk_config_group() -> None:
    """Deal Desk config helpers."""


@dealdesk_config_group.command("init")
@click.option(
    "--output",
    "output_path",
    default=Path("dealdesk.yaml"),
    type=click.Path(dir_okay=False, path_type=Path),
    help="Output path (default: ./dealdesk.yaml).",
)
def dealdesk_config_init(output_path: Path) -> None:
    """Write the default Deal Desk config to disk."""
    from g_gremlin.config import get_config_dir
    default_path = get_config_dir() / "dealdesk" / "dealdesk.yaml"
    if not default_path.exists():
        click.echo("Default Deal Desk config not found.", err=True)
        sys.exit(EXIT_ISSUES)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(default_path.read_text(encoding="utf-8"), encoding="utf-8")
    click.echo(f"Wrote default config to {output_path}")


__all__ = ["dealdesk_group"]



import pytest
import pandas as pd
import numpy as np
import math
from unittest.mock import patch
from pathlib import Path
from fmatch.core.engine import _init_score_cache_db, cached_score, apply_custom_rules
from fmatch.core import engine

# It's good to clear the ruleset cache between test runs if tests might modify it or rely on its state.
@pytest.fixture(autouse=True)
def clear_ruleset_cache():
    if hasattr(engine, "_RULESET_CACHE"):
        engine._RULESET_CACHE.clear()

# --- Tests for calculate_raw_signals ---
def test_calculate_raw_signals_all_identical_keys_dedupe():
    # Arrange
    key_value = "aaa"
    n_records = 100
    key_length = 3
    src_keys = pd.Series([key_value] * n_records)
    mode = engine.BlockingMode.DEDUPE

    expected_signals = {
        "RR": 0.0,
        "H": 0.0,
        "G": 0.0, # Perfect equality (all in one block)
        "P1": 1.0,
        "S1": 0.0,
        "k": float(key_length)
    }

    # Act
    signals = engine.calculate_raw_signals(src_keys, key_length, mode)

    # Assert
    assert signals is not None
    for key, val in expected_signals.items():
        assert signals[key] == pytest.approx(val)

def test_calculate_raw_signals_all_unique_keys_dedupe():
    # Arrange
    n_records = 100
    key_length = 3
    src_keys = pd.Series([f"key_{i:03}" for i in range(n_records)])
    mode = engine.BlockingMode.DEDUPE

    expected_signals = {
        "RR": 1.0,
        "H": 1.0,
        "G": 0.0,
        "P1": 1.0 / n_records,
        "S1": 1.0,
        "k": float(key_length)
    }

    # Act
    signals = engine.calculate_raw_signals(src_keys, key_length, mode)

    # Assert
    assert signals is not None
    for key, val in expected_signals.items():
        assert signals[key] == pytest.approx(val)

# --- Tests for normalize_signals ---
NORMALIZE_SCENARIOS = [
    ("Hn_zero", {"H": 0.0}, {"Hn": 0.0}),
    ("Hn_mid", {"H": 0.5}, {"Hn": 0.5}),
    ("Hn_one", {"H": 1.0}, {"Hn": 1.0}),
    ("Hn_missing", {}, {"Hn": 0.0}),
    ("Hn_outside_bounds_high", {"H": 1.5}, {"Hn": 1.0}),
    ("Hn_outside_bounds_low", {"H": -0.5}, {"Hn": 0.0}),
    ("Gn_G_zero", {"G": 0.0}, {"Gn": 1.0}),
    ("Gn_G_one", {"G": 1.0}, {"Gn": 0.0}),
    ("Gn_G_mid", {"G": 0.4}, {"Gn": 0.6}),
    ("Gn_missing", {}, {"Gn": 1.0}),
    ("Gn_G_outside_bounds_high", {"G": 1.5}, {"Gn": 0.0}),
    ("Gn_G_outside_bounds_low", {"G": -0.5}, {"Gn": 1.0}),
    ("RRn_RR_zero", {"RR": 0.0}, {"RRn": 0.0}),
    ("RRn_RR_one", {"RR": 1.0}, {"RRn": 1.0}),
    ("RRn_RR_mid", {"RR": 0.25}, {"RRn": 0.5}),
    ("RRn_missing", {}, {"RRn": 0.0}),
    ("RRn_RR_outside_bounds_high", {"RR": 1.5}, {"RRn": 1.0}),
    ("RRn_RR_outside_bounds_low", {"RR": -0.5}, {"RRn": 0.0}),
    ("P1n_P1_zero", {"P1": 0.0}, {"P1n": 1.0}),
    ("P1n_P1_at_threshold_edge_low", {"P1": 0.05}, {"P1n": 0.75}),
    ("P1n_P1_at_threshold", {"P1": 0.20}, {"P1n": 0.0}),
    ("P1n_P1_above_threshold", {"P1": 0.30}, {"P1n": 0.0}),
    ("P1n_P1_one", {"P1": 1.0}, {"P1n": 0.0}),
    ("P1n_missing", {}, {"P1n": 1.0}),
    ("P1n_P1_outside_bounds_high", {"P1": 1.5}, {"P1n": 0.0}),
    ("P1n_P1_outside_bounds_low", {"P1": -0.5}, {"P1n": 1.0}),
    ("S1n_S1_sweet_spot", {"S1": 0.10}, {"S1n": 1.0}),
    ("S1n_S1_zero", {"S1": 0.0}, {"S1n": 0.0}),
    ("S1n_S1_double_sweet_spot_boundary", {"S1": 0.20}, {"S1n": 0.0}),
    ("S1n_S1_mid_lower", {"S1": 0.05}, {"S1n": 0.5}),
    ("S1n_S1_mid_upper", {"S1": 0.15}, {"S1n": 0.5}),
    ("S1n_S1_far_high", {"S1": 0.50}, {"S1n": 0.0}),
    ("S1n_S1_far_low_clamped_raw", {"S1": -0.30}, {"S1n": 0.0}),
    ("S1n_missing", {}, {"S1n": 0.0}),
    ("S1n_S1_outside_bounds_high_clamped_raw", {"S1": 1.5}, {"S1n": 0.0}),
    ("S1n_S1_outside_bounds_low_clamped_raw", {"S1": -0.5}, {"S1n": 0.0}),
    ("kn_k_min", {"k": 2.0}, {"kn": 1.0}),
    ("kn_k_mid1", {"k": 3.0}, {"kn": 0.75}),
    ("kn_k_mid2", {"k": 4.0}, {"kn": 0.5}),
    ("kn_k_mid3", {"k": 5.0}, {"kn": 0.25}),
    ("kn_k_max", {"k": 6.0}, {"kn": 0.0}),
    ("kn_k_below_min_clamp", {"k": 1.0}, {"kn": 1.0}),
    ("kn_k_above_max_clamp", {"k": 7.0}, {"kn": 0.0}),
    ("kn_missing", {}, {"kn": 1.0}),
    ("combined_realistic", {"RR": 0.80, "H": 0.75, "G": 0.2, "P1": 0.15, "S1": 0.08, "k": 3.0},
     {"RRn": math.sqrt(0.80),"Hn": 0.75,"Gn": 0.8,"P1n": 0.25,"S1n": 0.8,"kn": 0.75}),
]

@pytest.mark.parametrize("test_name, raw_signals_input, expected_normalized_output", NORMALIZE_SCENARIOS, ids=[s[0] for s in NORMALIZE_SCENARIOS])
def test_normalize_signals(test_name, raw_signals_input, expected_normalized_output):
    normalized_result = engine.normalize_signals(raw_signals_input, "alnum")
    assert normalized_result is not None, f"{test_name} should not return None"
    for key, expected_value in expected_normalized_output.items():
        assert key in normalized_result, f"{test_name}: Expected key '{key}' not in result"
        assert normalized_result[key] == pytest.approx(expected_value, abs=1e-6), f"{test_name}: Mismatch for '{key}'"

# --- Tests for build_key ---
BUILD_KEY_VARIANT_SCENARIOS = [
    ("alnum_simple", ["Acme Inc."], "alnum", 3, "none", ["acm"]),
    ("alnum_with_punctuation", ["Test & Co."], "alnum", 4, "none", ["test"]),
    ("alnum_long_k", ["LongStringValue"], "alnum", 10, "none", ["longstring"]),
    ("alnum_short_k", ["Short"], "alnum", 3, "none", ["sho"]),
    ("alnum_numbers", ["123 Company St."], "alnum", 15, "none", ["123companyst"]),
    ("alnum_empty_input", [""], "alnum", 3, "none", []),
    ("alnum_all_punctuation", ["!@#$%^"], "alnum", 3, "none", []),
    ("alnum_unicode", ["Über Ällës"], "alnum", 6, "none", ["überäl"]),
    ("soundex_robert_k3", ["Robert"], "soundex", 3, "none", ["R163"]),
    ("soundex_robert_k4", ["Robert"], "soundex", 4, "none", ["R163"]),
    ("soundex_rupert_k4", ["Rupert"], "soundex", 4, "none", ["R163"]),
    ("soundex_ashcraft_k4", ["Ashcraft"], "soundex", 4, "none", ["A261"]),
    ("soundex_empty_input", [""], "soundex", 4, "none", []),
    ("soundex_numbers_only", ["12345"], "soundex", 4, "none", []),
    ("soundex_short_name", ["Bo"], "soundex", 4, "none", ["B000"]),
    ("k_len_zero_becomes_one", ["Test"], "alnum", 0, "none", ["t"]),
    ("k_len_negative_becomes_one", ["Test"], "alnum", -2, "none", ["t"]),
]

@pytest.mark.parametrize("test_name, input_list, variant, k_len, ruleset_name, expected_list", BUILD_KEY_VARIANT_SCENARIOS, ids=[s[0] for s in BUILD_KEY_VARIANT_SCENARIOS])
def test_build_key_variants(test_name, input_list, variant, k_len, ruleset_name, expected_list):
    input_series = pd.Series(input_list, dtype="string", name="test_col")
    expected_series = pd.Series(expected_list, dtype="string", name="test_col")
    result_series, _ = engine.build_key(input_series, variant, k_len, ruleset_name)
    pd.testing.assert_series_equal(result_series, expected_series, check_names=False)


def test_soundex_unicode_and_special_chars():
    """
    Tests that the soundex implementation correctly handles non-ASCII
    and special characters, producing a standard code.
    """
    # Arrange
    input_data = ["Élizabeth", "Джеймс", "O'Malley"]
    # Update the expectation for the transliterated Cyrillic string
    expected_codes = ["E421", "D252", "O540"]
    input_series = pd.Series(input_data)

    # Act
    result_codes = input_series.apply(engine.soundex).tolist()

    # Assert
    assert result_codes == expected_codes, "Soundex did not handle Unicode/special chars as expected"


# --- Tests for early_reject ---
EARLY_REJECT_SCENARIOS = [
    ("P1_reject_dominant_block", {"P1": 0.51, "RR": 0.5, "H": 0.5}, True),
    ("P1_borderline_no_reject", {"P1": 0.50, "RR": 0.5, "H": 0.5}, False),
    ("P1_safe_value", {"P1": 0.20, "RR": 0.5, "H": 0.5}, False),
    ("RR_H_reject_fragmentation", {"P1": 0.1, "RR": 0.9995, "H": 0.05}, True),
    ("RR_H_borderline_RR_no_reject", {"P1": 0.1, "RR": 0.999, "H": 0.05}, True),
    ("RR_H_borderline_H_no_reject", {"P1": 0.1, "RR": 0.9995, "H": 0.1}, True),
    ("RR_H_safe_RR", {"P1": 0.1, "RR": 0.900, "H": 0.05}, True),
    ("RR_H_safe_H", {"P1": 0.1, "RR": 0.9995, "H": 0.5}, False),
    ("no_rejection_conditions_met", {"P1": 0.3, "RR": 0.90, "H": 0.6}, False),
    ("missing_P1_key", {"RR": 0.5, "H": 0.5}, False),
    ("missing_RR_key", {"P1": 0.1, "H": 0.5}, False),
    ("missing_H_key", {"P1": 0.1, "RR": 0.9995}, False),
    ("missing_all_relevant_keys", {}, False),
    ("P1_reject_overrides_safe_RR_H", {"P1": 0.6, "RR": 0.90, "H": 0.6}, True),
]

@pytest.mark.parametrize("test_name, raw_signals_input, expected_rejection", EARLY_REJECT_SCENARIOS, ids=[s[0] for s in EARLY_REJECT_SCENARIOS])
def test_early_reject(test_name, raw_signals_input, expected_rejection):
    should_reject = engine.early_reject(raw_signals_input)
    assert should_reject == expected_rejection, f"{test_name}: Rejection status mismatch"

# --- Tests for determine_optimal_blocking_strategy ---
def test_determine_optimal_blocking_strategy_successful_selection_dedupe():
    # Arrange
    np.random.seed(42)
    data = {
        'id': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        'name': ["Apple Inc", "Apple Inc.", "Appel Corp", "Banana Ltd", "Banana Ltd", "Bonanza Co", "Cherry Plc", "Cherry Plc", "Cheri Co", "Date Co", "Date Co", "Date Co"],
        'city': ["London", "London", "Paris", "London", "Paris", "Paris", "Berlin", "Berlin", "Tokyo", "Tokyo", "Rome", "Rome"],
        'zip': [str(i) for i in range(100, 112)]
    }
    df_src_data = pd.DataFrame(data)
    mappings = [{"source": "name"}, {"source": "city"}, {"source": "zip"}]

    restrict_to_pairs_data = [(m['source'], m['source']) for m in mappings]

    original_hard_cap, engine.HARD_CAP_COMPS = engine.HARD_CAP_COMPS, 50
    original_min_qs, engine.MIN_ACCEPTABLE_QS = engine.MIN_ACCEPTABLE_QS, 0.1

    try:
        # FIX: Removed the unexpected 'key_len_range' keyword argument to
        # match the function's actual signature in engine.py.
        result = engine.determine_optimal_blocking_strategy(
            src_df=df_src_data,
            ref_df=None,
            restrict_to_pairs=restrict_to_pairs_data,
            preprocessing_variants=("alnum",)
        )
        assert result is not None, "Strategy should not be None"
        assert result.get("apply_blocking") is True
        assert result.get("is_fallback") is False
        assert result.get("source_block_col") == "name"
    finally:
        engine.HARD_CAP_COMPS, engine.MIN_ACCEPTABLE_QS = original_hard_cap, original_min_qs

# --- Recalculated test data for quality_score ---
NEW_WEIGHTS = {'rr': 0.10, 'h': 0.20, 'g': 0.30, 'p1': 0.25, 's1': 0.10, 'k': 0.05}
expected_all_one = sum(NEW_WEIGHTS.values()) * 0.1 # Penalized for RRn > 0.999
expected_only_rr_one = NEW_WEIGHTS['rr'] * 0.1 # Penalized for RRn > 0.999
expected_mixed = ((NEW_WEIGHTS['rr'] * 0.8) + (NEW_WEIGHTS['h'] * 0.7**2) + (NEW_WEIGHTS['g'] * 0.9) + (NEW_WEIGHTS['p1'] * 0.6) + (NEW_WEIGHTS['s1'] * 0.5) + (NEW_WEIGHTS['k'] * 0.75))
# Correct calculation: score is calculated first, then penalized.
# score = (0.10 * 1.0) + (0.20 * 1.0**2) = 0.30. Then, 0.30 * 0.1 (penalty) = 0.03
expected_some_missing = (NEW_WEIGHTS['rr'] * 1.0 + NEW_WEIGHTS['h'] * 1.0**2) * 0.1

QS_SCENARIOS = [
    ("all_normalized_zero", {"RRn": 0.0, "Hn": 0.0, "Gn": 0.0, "P1n": 0.0, "S1n": 0.0, "kn": 0.0}, 0.0),
    ("all_normalized_one", {"RRn": 1.0, "Hn": 1.0, "Gn": 1.0, "P1n": 1.0, "S1n": 1.0, "kn": 1.0}, expected_all_one),
    ("only_RRn_one", {"RRn": 1.0, "Hn": 0.0, "Gn": 0.0, "P1n": 0.0, "S1n": 0.0, "kn": 0.0}, expected_only_rr_one),
    ("only_Hn_one", {"RRn": 0.0, "Hn": 1.0, "Gn": 0.0, "P1n": 0.0, "S1n": 0.0, "kn": 0.0}, NEW_WEIGHTS['h']),
    ("mixed_values_realistic", {"RRn": 0.8, "Hn": 0.7, "Gn": 0.9, "P1n": 0.6, "S1n": 0.5, "kn": 0.75}, expected_mixed),
    ("some_keys_missing", {"RRn": 1.0, "Hn": 1.0}, expected_some_missing),
    ("all_keys_missing", {}, 0.0)
]

@pytest.mark.parametrize("test_name, normalized_signals_input, expected_qs", QS_SCENARIOS, ids=[s[0] for s in QS_SCENARIOS])
def test_quality_score(test_name, normalized_signals_input, expected_qs):
    calculated_qs = engine.quality_score(normalized_signals_input)
    assert calculated_qs == pytest.approx(expected_qs, abs=1e-6), f"{test_name}: Quality score mismatch"

# --- Test for cache roundtrip ---
def test_cached_score_roundtrip(tmp_path: Path):
    db_path = ":memory:"
    _init_score_cache_db(str(db_path))
    score1 = cached_score(s1="apple", s2="apply", algo="WRatio", processing_mode="scale", ruleset_name="test_ruleset", db_path=str(db_path))
    assert isinstance(score1, int) and score1 > 0
    score2 = cached_score(s1="apple", s2="apply", algo="WRatio", processing_mode="scale", ruleset_name="test_ruleset", db_path=str(db_path))
    assert score2 == score1

def test_apply_custom_rules_vectorized():
    """
    Tests that custom rules are applied correctly.
    """
    # Arrange
    df = pd.DataFrame({
        'score_levenshtein': [80, 95, 70],
        'is_exact_match': [False, True, False]
    })
    rules = [
        {"condition": "score_levenshtein > 85", "score_adjustment": 10},
        {"condition": "is_exact_match == True", "score_adjustment": 50}
    ]

    # Act
    result_df = apply_custom_rules(df, rules)

    # Assert
    assert "custom_rule_score" in result_df.columns
    assert result_df.loc[0, 'custom_rule_score'] == 0
    assert result_df.loc[1, 'custom_rule_score'] == 60
    assert result_df.loc[2, 'custom_rule_score'] == 0


def test_apply_custom_rules_security_no_side_effects(monkeypatch):
    """
    Tests that a malicious rule is not executed and has no effect,
    which aligns with the engine's behavior of logging errors instead of raising them.
    """
    # Arrange: Mock os.system to track if it gets called
    mock_system = patch('os.system').start()

    df = pd.DataFrame({'score': [80]})
    # This rule attempts to use a forbidden import and function
    malicious_rule = [{"condition": "__import__('os').system('echo pwned')", "score_adjustment": 1000}]

    # Act
    result_df = apply_custom_rules(df, malicious_rule)

    # Assert: The malicious function was NOT called, and no score was applied
    mock_system.assert_not_called()
    assert result_df['custom_rule_score'].iloc[0] == 0

    patch.stopall() # Clean up the patch
